# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "06/03/2020"


from silx.gui import qt

import tomwer.core.process.control.datalistener.datalistener
from orangecontrib.tomwer.widgets.utils import WidgetLongProcessing
from orangewidget import widget, gui
from orangewidget.widget import Output
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.blissscan import BlissScan
from tomwer.core.process.control.datalistener import DataListener
from tomwer.utils import docstring
from tomwer.synctools.datalistener import DataListenerQThread
from tomwer.gui.control.datalistener import DataListenerWidget
from tomwer.synctools.datalistener import MockDataListenerQThread
from tomwer.synctools.stacks.control.datalistener import DataListenerProcessStack
from tomwer.core.process.control.nxtomomill import NxTomomillProcess
from tomwer.core.process.control.datalistener.rpcserver import (
    send_signal_to_local_rpc_servers,
)
from orangewidget import settings

try:
    # new HDF5Config class
    from nxtomomill.io.config import TomoHDF5Config as HDF5Config
except ImportError:
    from nxtomomill.io.config import HDF5Config
from typing import Iterable
import functools
import logging
import signal
import sys

logger = logging.getLogger(__name__)


class DataListenerOW(
    widget.OWBaseWidget, WidgetLongProcessing, DataListener, openclass=True
):
    """
    This widget is used to listen to a server notifying the widget when an
    acquisition is finished.
    Then the bliss file will be converted to .nx file, NXtomo compliant.
    """

    name = "data listener"
    id = "orangecontrib.widgets.tomwer.control.DataListenerOW.DataListenerOW"
    description = (
        "The widget will receive information from bliss acquisition "
        "and wait for acquisition to be finished. Once finished it "
        "will call nxtomomill to convert from bliss .hdf5 to "
        "NXtomo compliant .nx file"
    )
    icon = "icons/datalistener.svg"
    priority = 10
    keywords = [
        "tomography",
        "file",
        "tomwer",
        "listener",
        "datalistener",
        "hdf5",
        "NXtomo",
    ]

    allows_cycle = True
    compress_signal = False

    want_main_area = True
    resizing_enabled = True

    ewoksorangeclass = (
        tomwer.core.process.control.datalistener.datalistener._DataListenerTaskPlaceHolder
    )

    _blissConfiguration = settings.Setting(dict())
    # to keep backward compatibility

    _nxtomo_cfg_file = settings.Setting(str())
    # to keep backward compatibility

    _static_input = settings.Setting(dict())

    class Outputs:
        data = Output(name="data", type=TomwerScanBase, doc="one scan to be process")

    def __init__(self, parent=None):
        widget.OWBaseWidget.__init__(self, parent)
        WidgetLongProcessing.__init__(self)
        DataListener.__init__(self)
        self._processingStack = DataListenerProcessStack()
        self._processingStack.sigComputationEnded.connect(self._signal_scan_ready)
        self._widget = DataListenerWidget(parent=self)
        self._mock = False

        self._box = gui.vBox(self.mainArea, self.name)
        layout = self._box.layout()
        layout.addWidget(self._widget)

        # signal / slot connection
        self._widget.sigActivate.connect(self._activated)
        self._widget.sigDeactivate.connect(self._deactivated)
        self._widget.sigConfigurationChanged.connect(self._jsonRPCConfigChanged)
        self._widget.sigCFGFileChanged.connect(self._nxtomoFileChanged)
        self._widget.sigAcquisitionEnded.connect(self._process_bliss_file_frm_tuple)
        self._widget.sigServerStopped.connect(self._serverStopped)

        # manage server stop when delete directly the widget or stop by Ctr+C
        signal.signal(signal.SIGINT, self.handleSigTerm)
        onDestroy = functools.partial(self._stopServerBeforeclosing)
        self.destroyed.connect(onDestroy)

        # set up
        self._loadSettings()

    def _loadSettings(self):
        if "bliss_server_configuration" in self._static_input:
            bliss_configuration = self._static_input["bliss_server_configuration"]
        else:
            bliss_configuration = self._blissConfiguration
        if bliss_configuration != {}:
            self._widget.setBlissServerConfiguation(bliss_configuration)
        if "nxtomomill_cfg_file" in self._static_input:
            nxtomo_cfg_file = self._static_input["nxtomomill_cfg_file"]
        else:
            nxtomo_cfg_file = self._nxtomo_cfg_file
        self._widget.setCFGFilePath(nxtomo_cfg_file)
        if "output_dir" in self._static_input:
            self._widget.setOutputFolder(self._static_input["output_dir"])

    def getNXTomomillConfiguration(self):
        cfg_file = self._widget.getCFGFilePath()
        if cfg_file in (None, ""):
            config = HDF5Config()
        else:
            try:
                config = HDF5Config.from_cfg_file(cfg_file)
            except Exception:
                config = HDF5Config()
        return config

    def _process_bliss_file_frm_tuple(self, t):
        master_file, entry, proposal_file, saving_file, success = t
        bliss_scan = BlissScan(
            master_file=master_file,
            entry=str(entry) + ".1",
            proposal_file=proposal_file,
            saving_file=saving_file,
        )
        configuration = self.getNXTomomillConfiguration()
        # overwrite output file
        configuration.output_file = NxTomomillProcess.deduce_output_file_path(
            bliss_scan.master_file,
            entry=bliss_scan.entry,
            outputdir=self._widget.getOutputFolder(),
        )
        if success:
            self._processingStack.add(scan=bliss_scan, configuration=configuration)
        else:
            # TODO: maybe remove the acquisition ?
            pass

    def _activated(self):
        self.activate(True)

    def _deactivated(self):
        self.activate(False)

    def _serverStopped(self):
        """
        Callback when the server is stopped
        """
        self.activate(False)

    @docstring(DataListener.activate)
    def activate(self, activate=True):
        if activate and not self.is_port_available():
            old = self._widget.blockSignals(True)
            self._widget.activate(activate=False)
            self._widget.blockSignals(old)
            dialog = _PortOccupyDialog(parent=self, port=self.port, host=self.host)
            dialog.setModal(False)
            if dialog.exec_() == qt.QDialog.Accepted:
                if dialog.retry_connection:
                    return self.activate(activate=True)
            else:
                return

        old = self._widget.blockSignals(True)
        self.set_configuration(self._widget.getBlissServerConfiguration())
        self._widget.activate(activate=activate)
        DataListener.activate(self, activate=activate)
        self.processing_state(activate, info="listener active")
        self._widget.blockSignals(old)

    def _signal_scan_ready(self, scan, future_scan):
        if scan is None:
            return
        assert isinstance(scan, Iterable)
        for s in scan:
            assert isinstance(s, TomwerScanBase)
            self.Outputs.data.send(s)

    def _ask_user_for_overwritting(self, file_path):
        msg = qt.QMessageBox(self)
        msg.setIcon(qt.QMessageBox.Question)
        types = qt.QMessageBox.Ok | qt.QMessageBox.Cancel
        msg.setStandardButtons(types)

        text = "NXtomomill will overwrite: \n %s. Do you agree ?" % file_path
        msg.setText(text)
        return msg.exec_() == qt.QMessageBox.Ok

    def _jsonRPCConfigChanged(self):
        self._blissConfiguration = self._widget.getBlissServerConfiguration()
        self._static_input[
            "bliss_server_configuration"
        ] = self._widget.getBlissServerConfiguration()
        self._static_input["output_dir"] = self._widget.getOutputFolder()
        if self.is_active():
            self.activate(False)
            self.activate(True)

    def _nxtomoFileChanged(self, cfg_file):
        self._nxtomo_cfg_file = cfg_file
        self._static_input["nxtomomill_cfg_file"] = cfg_file

    def setMock(self, mock, acquisitions):
        self._mock = mock
        self._mock_acquisitions = acquisitions

    @docstring(DataListenerWidget.getHost)
    def getHost(self):
        return self._widget.getHost()

    @docstring(DataListenerWidget.getPort)
    def getPort(self):
        return self._widget.getPort()

    def close(self):
        signal.signal(signal.SIGINT, signal.SIG_DFL)
        self.activate(False)
        super().close()

    @docstring(DataListener.create_listening_thread)
    def create_listening_thread(self):
        if self._mock is True:
            thread = MockDataListenerQThread(
                host=self.getHost(),
                port=self.getPort(),
                acquisitions=None,
                mock_acquisitions=self._mock_acquisitions,
            )
        else:
            thread = DataListenerQThread(
                host=self.getHost(), port=self.getPort(), acquisitions=None
            )
        # connect thread
        thread.sigAcquisitionStarted.connect(
            self._widget._acquisitionStarted, qt.Qt.DirectConnection
        )
        thread.sigAcquisitionEnded.connect(
            self._widget._acquisitionEnded, qt.Qt.DirectConnection
        )
        thread.sigScanAdded.connect(
            self._widget._acquisitionUpdated, qt.Qt.DirectConnection
        )
        thread.sigServerStop.connect(
            self._widget._serverStopped, qt.Qt.DirectConnection
        )
        return thread

    @docstring(DataListener.delete_listening_thread)
    def delete_listening_thread(self):
        self._listening_thread.sigAcquisitionStarted.disconnect(
            self._widget._acquisitionStarted
        )
        self._listening_thread.sigAcquisitionEnded.disconnect(
            self._widget._acquisitionEnded
        )
        self._listening_thread.sigScanAdded.disconnect(self._widget._acquisitionUpdated)
        self._listening_thread.sigServerStop.disconnect(self._widget._serverStopped)
        DataListener.delete_listening_thread(self)

    def _stopServerBeforeclosing(self):
        self.activate(False)

    def handleSigTerm(self, signo, *args, **kwargs):
        if signo == signal.SIGINT:
            self._stopServerBeforeclosing()
            sys.exit()

    def _get_n_scan_observe(self):
        return self._widget._observationWidget.observationTable.model().rowCount()

    def _get_n_scan_finished(self):
        return self._widget._historyWindow.scanHistory.model().rowCount()


class _PortOccupyDialog(qt.QDialog):
    def __init__(self, parent, port, host):
        qt.QDialog.__init__(self, parent)
        self.setLayout(qt.QVBoxLayout())
        self._retry = False
        self.port = port
        self.host = host

        mess = (
            "port ({port}) of {host} already in use. \n Maybe an other "
            "instance of `datalistener` is running in this session or "
            "another tomwer session. \n As this widget is connecting with "
            "bliss we enforce it to be unique.".format(port=port, host=host)
        )
        self.layout().addWidget(qt.QLabel(mess, self))
        self.setWindowTitle("Unable to launch two listener in parallel")

        types = qt.QDialogButtonBox.Cancel | qt.QDialogButtonBox.Retry
        self._buttons = qt.QDialogButtonBox(self)
        self._buttons.setStandardButtons(types)
        self.layout().addWidget(self._buttons)

        self._sendSIGTERMMsb = qt.QPushButton("send SIGTERM", self)
        self._sendSIGTERMMsb.setToolTip(
            "Try to send SIGTERM signal to the "
            "local tomwer-rpcserver if any "
            "occupies the reserved port"
        )
        self._buttons.addButton(self._sendSIGTERMMsb, qt.QDialogButtonBox.ActionRole)
        self._sendSIGKILLMsb = qt.QPushButton("send SIGKILL", self)
        self._sendSIGKILLMsb.setToolTip(
            "Try to send SIGKILL signal to the "
            "local tomwer-rpcserver if any "
            "occupies the reserved port"
        )
        self._buttons.addButton(self._sendSIGKILLMsb, qt.QDialogButtonBox.ActionRole)

        # set up
        # for now we don't want to show "send signal" feature
        self._sendSIGTERMMsb.hide()
        self._sendSIGKILLMsb.hide()

        # connect signal / slot
        self._buttons.button(qt.QDialogButtonBox.Cancel).released.connect(self.reject)
        self._buttons.button(qt.QDialogButtonBox.Retry).released.connect(
            self._retry_connect
        )
        self._sendSIGTERMMsb.released.connect(self._emitSigterm)
        self._sendSIGKILLMsb.released.connect(self._emitSigkill)

    @property
    def retry_connection(self):
        return self._retry

    def _emitSigterm(self, *args, **kwargs):
        send_signal_to_local_rpc_servers(signal.SIGTERM, port=self.port)

    def _emitSigkill(self, *args, **kwargs):
        send_signal_to_local_rpc_servers(signal.SIGKILL, port=self.port)

    def _retry_connect(self, *args, **kargs):
        self._retry = True
        self.accept()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "28/10/2020"


import unittest
import pytest
from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt
from tomwer.test.utils import skip_gui_test
from tomwer.core.utils.scanutils import MockHDF5
from tomwer.gui.edit.dkrfpatch import DarkRefPatchWidget
from tomwer.gui.edit.dkrfpatch import _DarkOrFlatUrl
from silx.io.url import DataUrl
from silx.io.utils import get_data
from tomoscan.esrf.hdf5scan import ImageKey
from nxtomomill.utils import add_dark_flat_nx_file
from tomwer.core.scan.hdf5scan import HDF5TomoScan
import tempfile
import shutil
import numpy


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestDarkOrFlatUrl(TestCaseQt):
    """
    Test some critical utils of the _DarkOrFlatUrl widget
    """

    def setUp(self):
        TestCaseQt.setUp(self)
        self._widgetForDark = _DarkOrFlatUrl(
            parent=None, type_=ImageKey.DARK_FIELD, when="start"
        )
        self._widgetForFlat = _DarkOrFlatUrl(
            parent=None, type_=ImageKey.FLAT_FIELD, when="start"
        )
        self.output_folder = tempfile.mkdtemp()
        #
        hdf5_mock = MockHDF5(
            scan_path=self.output_folder,
            n_ini_proj=20,
            n_proj=20,
            create_ini_dark=True,
            create_ini_ref=True,
            create_final_ref=True,
        )
        self._scan = hdf5_mock.scan
        self.scan_url = DataUrl(
            file_path=self._scan.master_file, data_path=self._scan.entry
        )

    def tearDown(self):
        shutil.rmtree(self.output_folder)
        self._widgetForDark.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._widgetForDark.close()
        self._widgetForDark = None
        self._widgetForFlat.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._widgetForFlat.close()
        self._widgetForFlat = None
        TestCaseQt.tearDown(self)

    def testSetUrlForFlat(self):
        self._widgetForFlat.setUrl(self.scan_url)
        self.assertEqual(self._widgetForFlat._optionsCB.count(), 3)
        data_url = self._widgetForFlat._redirectDataPath(self.scan_url)
        image_keys = self._widgetForFlat._getImageKey(data_url)
        self.assertTrue(image_keys is not None)
        slices_index_0 = self._widgetForFlat._getSlices(image_keys, 0)
        self.assertTrue(slices_index_0 is not None)
        self.assertEqual(
            list(range(slices_index_0.start, slices_index_0.stop)),
            [1, 2, 3, 4, 5, 6, 7, 8, 9, 10],
        )
        slices_index_1 = self._widgetForFlat._getSlices(image_keys, 1)
        self.assertTrue(slices_index_1 is not None)
        self.assertEqual(slices_index_1, slice(31, 41))

    def testSetUrlForDark(self):
        self._widgetForDark.setUrl(self.scan_url)
        self._widgetForDark.show()
        self.assertEqual(self._widgetForDark._optionsCB.count(), 2)
        data_url = self._widgetForDark._redirectDataPath(self.scan_url)
        image_keys = self._widgetForDark._getImageKey(data_url)
        slices_index_0 = self._widgetForDark._getSlices(image_keys, 0)
        self.assertEqual(
            list(range(slices_index_0.start, slices_index_0.stop)),
            [
                0,
            ],
        )

        slices_index_1 = self._widgetForDark._getSlices(image_keys, 1)
        self.assertTrue(slices_index_1 is None)


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestDarkRefPatchWidget(TestCaseQt):
    """
    Simple test to insure DarkRefWidget is working
    """

    def setUp(self):
        TestCaseQt.setUp(self)
        self._widget = DarkRefPatchWidget(parent=None)
        self.output_folder1 = tempfile.mkdtemp()
        self.output_folder2 = tempfile.mkdtemp()

        hdf5_mock = MockHDF5(
            scan_path=self.output_folder1,
            n_ini_proj=20,
            n_proj=20,
            create_ini_dark=True,
            create_ini_ref=True,
            create_final_ref=True,
        )
        self._scanWithDarkAndRef = hdf5_mock.scan

        hdf5_mock = MockHDF5(
            scan_path=self.output_folder2,
            n_ini_proj=20,
            n_proj=20,
            create_ini_dark=False,
            create_ini_ref=False,
            create_final_ref=False,
        )
        self._scan = hdf5_mock.scan
        assert len(self._scan.darks) == 0
        assert len(self._scan.flats) == 0

    def tearDown(self):
        shutil.rmtree(self.output_folder1)
        shutil.rmtree(self.output_folder2)
        self._widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._widget.close()
        self._widget = None
        self.qapp.processEvents()
        TestCaseQt.tearDown(self)

    def testEdition(self):
        # insure the widget work and post processing made by the application
        # and the widget succeed to
        url = DataUrl(
            file_path=self._scanWithDarkAndRef.master_file,
            data_path=self._scanWithDarkAndRef.entry,
        )
        self._widget.setStartDarkUrl(url, serie_index=0)
        self._widget.setStartFlatUrl(url, serie_index=1)
        self.assertTrue(self._widget.getStartDarkUrl() is not None)
        self.assertTrue(self._widget.getStartDarkUrl().is_valid())
        self.assertTrue(self._widget.getStartFlatUrl() is not None)
        self.assertTrue(self._widget.getStartFlatUrl().is_valid())
        self.assertTrue(self._widget.getEndDarkUrl() is None)
        self.assertTrue(self._widget.getEndFlatUrl() is None)
        self.process()
        new_scan = HDF5TomoScan(scan=self._scan.master_file, entry=self._scan.entry)
        self.assertEqual(len(new_scan.darks), 1)
        self.assertTrue(0 in new_scan.darks)
        numpy.testing.assert_array_equal(
            get_data(new_scan.darks[0]),
            get_data(self._scanWithDarkAndRef.darks[0]),
        )

        self.assertEqual(len(self._scanWithDarkAndRef.flats), 20)
        self.assertEqual(len(new_scan.flats), 10)

        self.assertEqual(tuple(new_scan.flats.keys()), tuple(range(1, 11)))
        numpy.testing.assert_array_equal(
            get_data(new_scan.flats[1]),
            get_data(self._scanWithDarkAndRef.flats[31]),
        )
        numpy.testing.assert_array_equal(
            get_data(new_scan.flats[10]),
            get_data(self._scanWithDarkAndRef.flats[40]),
        )

    def process(self):
        url_sd = self._widget.getStartDarkUrl()
        url_sf = self._widget.getStartFlatUrl()
        url_ed = self._widget.getEndDarkUrl()
        url_ef = self._widget.getEndFlatUrl()
        add_dark_flat_nx_file(
            file_path=self._scan.master_file,
            entry=self._scan.entry,
            darks_start=url_sd,
            flats_start=url_sf,
            darks_end=url_ed,
            flats_end=url_ef,
        )


def suite():
    test_suite = unittest.TestSuite()
    for ui in (TestDarkRefPatchWidget, TestDarkOrFlatUrl):
        test_suite.addTest(unittest.defaultTestLoader.loadTestsFromTestCase(ui))
    return test_suite


if __name__ == "__main__":
    unittest.main(defaultTest="suite")



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["C. Nemoz", "H. Payno"]
__license__ = "MIT"
__date__ = "10/01/2018"


import logging
from orangewidget import widget, gui
from processview.gui.processmanager import ProcessManagerWindow

logger = logging.getLogger(__name__)


class AdvancementOW(widget.OWBaseWidget, openclass=True):
    """
    A simple widget managing the copy of an incoming folder to an other one

    :param parent: the parent widget
    """

    # note of this widget should be the one registred on the documentation
    name = "advancement"
    id = "orangecontrib.widgets.tomwer.control.AdvancementOW.AdvancementOW"
    description = "This widget can display advancement of processes and scans"
    icon = "icons/advancement.svg"
    priority = 5
    keywords = ["tomography", "process", "advancement"]

    want_main_area = True
    resizing_enabled = True
    compress_signal = False

    def __init__(self, parent=None):
        widget.OWBaseWidget.__init__(self, parent)

        self._widget = ProcessManagerWindow(parent=self)

        self._box = gui.vBox(self.mainArea, self.name)
        layout = self._box.layout()
        layout.addWidget(self._widget)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["C. Nemoz", "H. Payno"]
__license__ = "MIT"
__date__ = "25/10/2016"

import tomwer.core.process.control.datawatcher.datawatcher
from ..utils import WidgetLongProcessing
from orangewidget import widget, gui
from orangewidget.settings import Setting
from orangewidget.widget import Output
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.gui.control.datawatcher import DataWatcherWidget
from tomwer.web.client import OWClient
import functools
import logging

logger = logging.getLogger(__name__)


class DataWatcherOW(
    widget.OWBaseWidget, OWClient, WidgetLongProcessing, openclass=True
):
    """
    This widget is used to observe a selected folder and his sub-folders to
    detect if they are containing valid-finished acquisitions.
    """

    name = "data watcher"
    id = "orangecontrib.widgets.tomwer.datawatcherwidget.DataWatcherOW"
    description = (
        "The widget will observe folder and sub folders of a given"
        " path and waiting for acquisition to be ended."
        " The widget will infinitely wait until an acquisition is "
        "ended. If an acquisition is ended then a signal "
        "containing the folder path is emitted."
    )
    icon = "icons/datawatcher.svg"
    priority = 12
    keywords = ["tomography", "file", "tomwer", "observer", "datawatcher"]

    allows_cycle = True
    compress_signal = False

    want_main_area = True
    resizing_enabled = True

    folderObserved = Setting(str())

    DEFAULT_DIRECTORY = "/lbsram/data/visitor"

    ewokstaskclass = tomwer.core.process.control.datawatcher.datawatcher.DataWatcher

    class Outputs:
        data = Output(name="data", type=TomwerScanBase)

    def __init__(self, parent=None, displayAdvancement=True):
        """Simple class which will check advancement state of the acquisition
        for a specific folder

        :param parent: the parent widget
        """
        widget.OWBaseWidget.__init__(self, parent)
        OWClient.__init__(self)
        WidgetLongProcessing.__init__(self)
        self._widget = DataWatcherWidget(parent=self)
        self._widget.setFolderObserved(self.folderObserved)

        self._box = gui.vBox(self.mainArea, self.name)
        layout = self._box.layout()
        layout.addWidget(self._widget)

        # signal / slot connection
        self._widget.sigFolderObservedChanged.connect(self._updateSettings)
        self._widget.sigScanReady.connect(self._sendSignal)
        callback_start = functools.partial(self.processing_state, True, "watching on")
        self._widget.sigObservationStart.connect(callback_start)
        callback_end = functools.partial(self.processing_state, False, "watching off")
        self._widget.sigObservationEnd.connect(callback_end)

    def _updateSettings(self):
        self.folderObserved = self._widget.getFolderObserved()

    @property
    def currentStatus(self):
        return self._widget.currentStatus

    @property
    def sigTMStatusChanged(self):
        return self._widget.sigTMStatusChanged

    def resetStatus(self):
        self._widget.resetStatus()

    def _sendSignal(self, scan):
        assert isinstance(scan, TomwerScanBase)
        self.Outputs.data.send(scan)

    def setFolderObserved(self, path):
        self._widget.setFolderObserved(path)

    def setObservation(self, b):
        self._widget.setObservation(b)

    def setTimeBreak(self, val):
        """
        Set the time break between two loop observation
        :param val: time (in sec)
        """
        self._widget.setWaitTimeBtwLoop(val)

    def startObservation(self):
        try:
            self.processing_info("processing")
        except Exception:
            pass
        self._widget.start()

    def stopObservation(self, succes=False):
        self._widget.stop(succes)
        try:
            self.Processing.clear()
        except Exception:
            pass



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "10/01/2019"


import os
from silx.gui import qt
from tomwer.core.process.control.datawatcher.datawatcherobserver import (
    _OngoingObservation,
)
from tomwer.core.process.control.datawatcher.datawatcherobserver import (
    _DataWatcherObserver_MixIn,
)
from tomwer.core.process.control.datawatcher.datawatcherobserver import (
    _DataWatcherFixObserverMixIn,
)
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.process.control.datawatcher import status as datawatcherstatus
import logging
from tomwer.core.scan.scanfactory import ScanFactory

logger = logging.getLogger(__name__)


class _QDataWatcherObserver(_DataWatcherObserver_MixIn, qt.QThread):
    """
    DatWatcherObserver implementation with a qt.QThread

    We have two implementations in order to avoid hard dependancy on qt for
    tomwer.core package
    """

    sigScanReady = qt.Signal(TomwerScanBase)

    def __init__(
        self,
        obsMethod,
        observationClass,
        headDir=None,
        startByOldest=False,
        srcPattern=None,
        destPattern=None,
        ignoredFolders=None,
    ):
        qt.QThread.__init__(self)
        _DataWatcherObserver_MixIn.__init__(
            self,
            obsMethod=obsMethod,
            observationClass=observationClass,
            headDir=headDir,
            startByOldest=startByOldest,
            srcPattern=srcPattern,
            destPattern=destPattern,
            ignoredFolders=ignoredFolders,
            time_between_loops=0.2,
        )
        self.observations.sigScanReady.connect(self._signalScanReady)

    def _signalScanReady(self, scan):
        assert isinstance(scan, TomwerScanBase)
        self.sigScanReady.emit(scan)

    def _getObserver(self, scanID):
        return _QDataWatcherFixObserver(
            scanID=scanID,
            obsMethod=self.obsMethod,
            srcPattern=self.srcPattern,
            destPattern=self.destPattern,
            patternObs=self._patternObs,
            observationRegistry=self.observations,
        )

    def run(self):
        def process(directory):
            if (
                self.observations.isObserving(directory) is False
                and self.dataWatcherProcess._isScanDirectory(directory)
                and directory not in self.observations.ignoredFolders
            ):
                self.observe(directory)

            try:
                for f in os.listdir(directory):
                    if os.path.isdir(os.path.join(directory, f)):
                        process(os.path.join(directory, f))
            except Exception:
                pass

        if not os.path.isdir(self.headDir):
            logger.warning("can't observe %s, not a directory" % self.headDir)
            return
        self.dataWatcherProcess = self._getDataWatcherProcess()
        process(self.headDir)

        self._processObservation()

    def waitForObservationFinished(self, timeOut=10):
        threads = list(self.observations.dict.values())
        for thread in threads:
            thread.wait(timeOut)

    def wait(self, *args, **kwargs):
        self.waitForObservationFinished()
        super(_QDataWatcherObserver, self).wait(*args, **kwargs)


class _QOngoingObservation(_OngoingObservation, qt.QObject):
    """
    _OngoingObservation with a QObject and signals for each event
    """

    sigScanReady = qt.Signal(TomwerScanBase)
    """Emitted when a finished acquisition is detected"""
    sigObsAdded = qt.Signal(str)
    """Signal emitted when an observation is added"""
    sigObsRemoved = qt.Signal(str)
    """Signal emitted when an observation is removed"""
    sigObsStatusReceived = qt.Signal(str, str)
    """Signal emitted when receiving a new observation status"""

    def __init__(self):
        qt.QObject.__init__(self)
        _OngoingObservation.__init__(self)

    def _acquisition_ended(self, scanID):
        _OngoingObservation._acquisition_ended(self, scanID=scanID)
        try:
            scan = ScanFactory.create_scan_object(scan_path=scanID)
        except Exception as e:
            logger.error("Fail to create TomoBase instance from", scanID, "Error is", e)
        else:
            self.sigScanReady.emit(scan)

    def add(self, observer):
        already_observing = self.isObserving(observer.path)
        _OngoingObservation.add(self, observer=observer)
        if not already_observing:
            observer.sigStatusChanged.connect(self._updateStatus)
            self.sigObsAdded.emit(observer.path)

    def remove(self, observer):
        observing = self.isObserving(observer.path)
        _OngoingObservation.remove(self, observer=observer)
        if observing is True:
            observer.sigStatusChanged.disconnect(self._updateStatus)
            self.sigObsRemoved.emit(observer.path)

    def _updateStatus(self, status, scan):
        if self.isObserving(scan) is True:
            self.sigObsStatusReceived.emit(
                scan, datawatcherstatus.DICT_OBS_STATUS[status]
            )
        _OngoingObservation._updateStatus(self, status=status, scan=scan)

    def reset(self):
        # self.ignoreFolders = []
        for scanID, observer in self.dict:
            observer.sigStatusChanged.disconnect(self._updateStatus)
            observer.quit()
        self.dict = {}


class _QDataWatcherFixObserver(_DataWatcherFixObserverMixIn, qt.QThread):
    """
    Implementation of the _DataWatcherFixObserverMixIn with a qt.QThread
    """

    sigStatusChanged = qt.Signal(int, str)
    """signal emitted when the status for a specific directory change
    """

    def __init__(
        self,
        scanID,
        obsMethod,
        srcPattern,
        destPattern,
        patternObs,
        observationRegistry,
    ):
        qt.QThread.__init__(self)
        _DataWatcherFixObserverMixIn.__init__(
            self,
            scanID=scanID,
            obsMethod=obsMethod,
            srcPattern=srcPattern,
            destPattern=destPattern,
            patternObs=patternObs,
            observationRegistry=observationRegistry,
        )

    def run(self):
        if not os.path.isdir(self.path):
            logger.info("can't observe %s, not a directory" % self.path)
            self.status = "failure"
            self.sigStatusChanged.emit(
                datawatcherstatus.OBSERVATION_STATUS[self.status], self.path
            )
            self.validation = -1
            return

        try:
            scan = ScanFactory.create_scan_object(scan_path=self.path)
        except ValueError as e:
            logger.error(e)
        else:
            if scan.is_abort(
                src_pattern=self.srcPattern, dest_pattern=self.destPattern
            ):
                if self.status != "aborted":
                    logger.info("Acquisition %s has been aborted" % self.path)
                    self.dataWatcherProcess._removeAcquisition(
                        scanID=self.path, reason="acquisition aborted by the user"
                    )

                    self.status = "aborted"
                self.sigStatusChanged.emit(
                    datawatcherstatus.OBSERVATION_STATUS[self.status], self.path
                )
                self.validation = -2
                return
            dataComplete = self.dataWatcherProcess.is_data_complete()

            if dataComplete is True:
                self.status = "acquisition ended"
                self.sigStatusChanged.emit(
                    datawatcherstatus.OBSERVATION_STATUS[self.status], self.path
                )
                self.validation = 1
            else:
                self.status = "waiting for acquisition ending"
                self.sigStatusChanged.emit(
                    datawatcherstatus.OBSERVATION_STATUS[self.status], self.path
                )
                self.validation = 0
        return



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "16/06/2021"


import unittest
from tomwer.core.process.control.volumesymlink import VolumeSymbolicLinkProcess
from tomwer.core.utils.scanutils import MockEDF
from tomwer.core.scan.edfscan import EDFTomoScan
import tempfile
import shutil
import os


class TestVolumeSymbolicLink(unittest.TestCase):
    """Simple test of the VolumeSymbolicLink"""

    def setUp(self) -> None:
        self._scan_folder = tempfile.mkdtemp()
        self._output_folder = tempfile.mkdtemp()
        MockEDF.fastMockAcquisition(self._scan_folder)
        MockEDF.mockReconstruction(self._scan_folder, nRecons=2, nPagRecons=2)
        self._scan = EDFTomoScan(self._scan_folder)
        # here we simply estimate that the slice reconstruction will be
        # volume reconstruction but the mechanism is the same.
        self._scan.set_latest_vol_reconstructions(self._scan.get_reconstructions_urls())

    def tearDown(self) -> None:
        for folder in (self._scan_folder, self._output_folder):
            shutil.rmtree(folder)

    def test(self):
        """simple test that the creation of a symbolic link works"""
        vlink_process = VolumeSymbolicLinkProcess(inputs={"data": self._scan})
        vlink_process.set_configuration(
            {
                "output_type": "static",
                "output_folder": self._output_folder,
            }
        )
        self.assertEqual(len(os.listdir(self._output_folder)), 0)
        vlink_process.run()
        self.assertEqual(len(os.listdir(self._output_folder)), 4)

        self.assertTrue(isinstance(vlink_process.program_name(), str))
        self.assertTrue(isinstance(vlink_process.program_version(), str))
        self.assertTrue(isinstance(vlink_process.definition(), str))



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/


__authors__ = ["H.Payno"]
__license__ = "MIT"
__date__ = "09/08/2018"


import os
import re
from collections import OrderedDict
from math import ceil
import fabio
import numpy
import functools
import logging
from tomoscan.scanbase import FOV
from tomwer.core.process.reconstruction.darkref.settings import (
    REFHST_PREFIX,
    DARKHST_PREFIX,
)
from tomwer.core.progress import Progress
from silx.io.url import DataUrl
import silx.io.utils
from glob import glob
from .scanbase import TomwerScanBase
from tomoscan.esrf.edfscan import EDFTomoScan as _tsEDFTomoScan
from tomwer.utils import docstring
from processview.core.dataset import DatasetIdentifier
from typing import Optional

_logger = logging.getLogger(__name__)


global counter_rand
counter_rand = 1  # used to be sure to return a unique index on recons slices


class _EDFTomoScanIdentifier(DatasetIdentifier):
    def __init__(self, dataset, edf_folder):
        super().__init__(dataset)
        self.edf_folder = os.path.realpath(os.path.abspath(edf_folder))

    @docstring(DatasetIdentifier)
    def long_description(self) -> str:
        return self.edf_folder

    def __str__(self):
        return os.path.basename(self.edf_folder)

    def __eq__(self, other):
        if isinstance(other, _EDFTomoScanIdentifier):
            return self.edf_folder == other.edf_folder
        else:
            return False

    def __hash__(self):
        return hash(self.edf_folder)


class EDFTomoScan(_tsEDFTomoScan, TomwerScanBase):
    """
    Class used to represent a tomography acquisition with hdf5 files.

    :param Union[str, None] scan: path of the scan
    """

    def __init__(self, scan, overwrite_proc_file=False):
        _tsEDFTomoScan.__init__(self, scan=scan)
        TomwerScanBase.__init__(self)

        self._dark = None
        self._process_file = self.get_process_file_name(self)
        self._init_index_process_file(overwrite_proc_file=overwrite_proc_file)

        if scan is not None:
            self.update()

    @staticmethod
    def get_process_file_name(scan):
        if scan.path is not None:
            basename = os.path.basename(scan.path)
            basename = "_".join((basename, "tomwer_processes.h5"))
            return os.path.join(scan.path, basename)
        else:
            return None

    def clear_caches(self):
        _tsEDFTomoScan.clear_caches(self)
        TomwerScanBase.clear_caches(self)
        self._dark = None

    @staticmethod
    def directory_contains_scan(directory, src_pattern, dest_pattern):
        """
        Check if the given directory is holding an acquisition

        :param str directory: directory we want to check
        :param str src_pattern: buffer name pattern ('lbsram')
        :param dest_pattern: output pattern (''). Needed because some
                             acquisition can split the file produce between
                             two directories. This is the case for edf,
                             where .info file are generated in /data/dir
                             instead of /lbsram/data/dir
        :return: does the given directory contains any acquisition
        :rtype: bool
        """
        aux = directory.split(os.path.sep)
        info_name = os.path.join(directory, aux[-1] + EDFTomoScan.INFO_EXT)

        if src_pattern:
            info_name = info_name.replace(src_pattern, dest_pattern, 1)
        return os.path.isfile(info_name)

    def data_flat_field_correction(self, data, index=None):
        """
        Apply the flat field correction on the given data.

        :param numpy.ndarray data: radio to correct
        :return numpy.ndarray: corrected data
        """
        dark = self.getDark()
        flat = self.getFlat()
        return self._flat_field_correction(
            data=data,
            dark=dark,
            flat1=flat,
            flat2=None,
            index_flat1=-1,
            index_flat2=-1,
            index_proj=index,
        )

    def update(self):
        """update list of radio and reconstruction by parsing the scan folder"""
        _tsEDFTomoScan.update(self)
        self.reconstructions = EDFTomoScan.get_reconstructions_paths(self.path)

    def projections_with_angle(self):
        return self.get_proj_angle_url()

    @staticmethod
    def get_reconstructions_paths(scanID, withIndex=False):
        """
        Return the dict of files:
        * fitting with a reconstruction pattern and ending by .edf
        * .vol files

        :param scanID: is the path to the folder of acquisition
        :param bool withIndex: if False then return a list of slices otherwise
            return a dict with the index of the slice reconstructed.
        """

        def containsDigits(input):
            return any(char.isdigit() for char in input)

        if (scanID is None) or (not os.path.isdir(scanID)):
            if withIndex is True:
                return {}
            else:
                return []

        pyhst_files = TomwerScanBase.get_pyhst_recons_file(scanID)
        if pyhst_files is not None:
            return TomwerScanBase.getReconstructedFilesFromParFile(
                pyhst_files, with_index=withIndex
            )
        else:
            folderBasename = os.path.basename(scanID)
            files = {} if withIndex is True else []
            if os.path.isdir(scanID):
                for f in os.listdir(scanID):
                    if (
                        f.endswith(TomwerScanBase.VALID_RECONS_EXTENSION)
                        and f.startswith(folderBasename)
                        and "slice_" in f
                    ):
                        local_str = f
                        for extension in TomwerScanBase.VALID_RECONS_EXTENSION:
                            if local_str.endswith(extension):
                                local_str = local_str.rsplit(extension, 1)[0]
                        if "slice_" in local_str:
                            if "slice_pag_" in local_str:
                                indexStr = local_str.split("slice_pag_")[-1].split("_")[
                                    0
                                ]
                            else:
                                indexStr = local_str.split("slice_")[-1].split("_")[0]
                            if containsDigits(indexStr):
                                gfile = os.path.join(scanID, f)
                                assert os.path.isfile(gfile)
                                if withIndex is True:
                                    files[
                                        EDFTomoScan.get_index_reconstructed(f, scanID)
                                    ] = gfile
                                else:
                                    files.append(gfile)
                    if f.endswith(".vol"):
                        if withIndex is True:
                            files[
                                EDFTomoScan.get_index_reconstructed(f, scanID)
                            ] = os.path.join(scanID, f)
                        else:
                            files.append(os.path.join(scanID, f))
            return files

    def load_from_dict(self, desc):
        _tsEDFTomoScan.load_from_dict(self, desc)
        TomwerScanBase.load_from_dict(self, desc)
        return self

    def _get_scheme(self):
        """

        :return: scheme to read url
        :rtype: str
        """
        return "fabio"

    @docstring(TomwerScanBase.get_sinogram)
    @functools.lru_cache(maxsize=16, typed=True)
    def get_sinogram(self, line, subsampling=1):
        """

        extract the sinogram from projections

        :param line: which sinogram we want
        :type: int
        :param subsampling: subsampling to apply if any. Allows to skip some io
        :type: int
        :return: sinogram from the radio lines
        :rtype: numpy.array
        """
        _logger.info(
            "compute sinogram for line %s of %s (subsampling: %s)"
            % (line, self.path, subsampling)
        )
        assert isinstance(line, int)
        if self.tomo_n is not None and self.dim_2 is not None and line > self.dim_2:
            raise ValueError("requested line %s is not in the scan")
        else:
            y_dim = ceil(self.tomo_n / subsampling)
            sinogram = numpy.empty((y_dim, self.dim_1))
            proj_urls = self.get_proj_angle_url()
            assert len(proj_urls) >= self.tomo_n
            proj_sort = list(proj_urls.keys())
            proj_sort = list(filter(lambda x: not isinstance(x, str), proj_sort))
            proj_sort.sort()
            advancement = Progress(
                name="compute sinogram for %s, line=%s,"
                "sampling=%s" % (os.path.basename(self.path), line, subsampling)
            )
            advancement.setMaxAdvancement(len(proj_sort))
            for i_proj, proj in enumerate(proj_sort):
                if i_proj % subsampling == 0:
                    url = proj_urls[proj]
                    radio = silx.io.utils.get_data(url)
                    radio = self.data_flat_field_correction(radio)
                    sinogram[i_proj // subsampling] = radio[line]
                advancement.increaseAdvancement(1)
            return sinogram

    @functools.lru_cache(maxsize=3)
    def getFlat(self, index=None):
        """
        If projectionI is not requested then return the mean value. Otherwise
        return the interpolated value for the requested projection.

        :param Union[int,None] index: index of the projection for which we want
                                      the flat
        :return: Flat field value or None if can't deduce it
        """
        data = self._extractFromOneFile("refHST.edf", what="flat")
        if data is not None:
            return data

        data = self._extractFromPrefix(REFHST_PREFIX, what="flat", proI=index)
        if data is not None:
            return data

        _logger.warning("Cannot retrieve flat file from %s" % self.path)
        return None

    @docstring(TomwerScanBase.get_proj_angle_url)
    def get_proj_angle_url(self, use_cache: bool = True):
        if not use_cache:
            self._cache_proj_urls = None

        if self._cache_proj_urls is None:
            self._cache_proj_urls = _tsEDFTomoScan.get_proj_angle_url(self)
        return self._cache_proj_urls

    @functools.lru_cache()
    def getDark(self):
        """
        For now only deal with one existing dark file.

        :return: image of the dark if existing. Else None
        """
        if self._dark is None:
            # first try to retrieve data from dark.edf file or darkHST.edf files
            self._dark = self._extractFromOneFile("dark.edf", what="dark")
            if self._dark is None:
                self._dark = self._extractFromOneFile("darkHST.edf", what="dark")
            if self._dark is None:
                self._dark = self._extractFromPrefix(DARKHST_PREFIX, what="dark")
            if self._dark is None:
                self._dark = self._extractFromPrefix("darkend", what="dark")

            if self._dark is None:
                _logger.warning("Cannot retrieve dark file from %s" % self.path)

        return self._dark

    def _extractFromOneFile(self, f, what):
        if self.path is None:
            return None
        path = os.path.join(self.path, f)
        if os.path.exists(path):
            _logger.info("Getting %s from %s" % (what, f))
            try:
                data = fabio.open(path).data
            except Exception:
                return None
            else:
                if data.ndim == 2:
                    return data
                elif data.ndim == 3:
                    _logger.warning(
                        "%s file contains several images. Taking "
                        "the mean value" % what
                    )
                    return numpy.mean(data.ndim)
        else:
            return None

    @staticmethod
    def guess_index_frm_EDFFile_name(_file):
        name = _file
        if name.endswith(".edf"):
            name = name.rstrip(".edf")
        ic = []
        while name[-1].isdigit():
            ic.append(name[-1])
            name = name[:-1]

        if len(ic) == 0:
            return None
        else:
            orignalOrder = ic[::-1]
            return int("".join(orignalOrder))

    def _extractFromPrefix(self, pattern, what, proI=None):
        if self.path is None:
            return None
        files = glob(os.path.join(self.path, pattern + "*.edf"))
        if len(files) == 0:
            return None
        else:
            d = {}
            for f in files:
                index = self.guess_index_frm_EDFFile_name(f)
                if index is None:
                    _logger.error("cannot retrieve projection index for %s" "" % f)
                    return None
                else:
                    d[index] = fabio.open(f).data

            if len(files) == 1:
                return d[list(d.keys())[0]]

            oProj = OrderedDict(sorted(d.items()))
            # for now we only deal with interpolation between the higher
            # and the lower acquired file ()
            lowPI = list(oProj.keys())[0]
            uppPI = list(oProj.keys())[-1]

            lowPD = oProj[lowPI]
            uppPD = oProj[uppPI]

            if len(oProj) > 2:
                _logger.info(
                    "Only bordering projections (%s and %s) will "
                    "be used for extracting %s" % (lowPI, uppPI, what)
                )

            uppPI = uppPI
            index = proI
            if index is None:
                index = (uppPI - lowPI) / 2

            if (index >= lowPI) is False:
                index = lowPI
                _logger.warning(
                    "ProjectionI not in the files indexes"
                    "range (projectionI >= lowerProjIndex)"
                )

            if (index <= uppPI) is False:
                index = uppPI
                _logger.warning(
                    "ProjectionI not in the files indexes"
                    "range upperProjIndex >= projectionI"
                )

            # simple interpolation
            _nRef = uppPI - lowPI
            lowPI = lowPI

            w0 = (lowPI + (uppPI - index)) / _nRef
            w1 = index / _nRef

            return w0 * lowPD + w1 * uppPD

    @staticmethod
    def get_index_reconstructed(reconstructionFile, scanID):
        """Return the slice reconstructed of a file from her name

        :param str reconstructionFile: the name of the file
        """
        folderBasename = os.path.basename(scanID)
        if reconstructionFile.endswith(".edf") and reconstructionFile.startswith(
            folderBasename
        ):
            localstring = reconstructionFile.rstrip(".edf")
            # remove the scan
            localstring = re.sub(folderBasename, "", localstring)
            s = localstring.split("_")
            if s[-1].isdigit():
                return int(s[-1])
            else:
                _logger.warning(
                    "Fail to find the slice reconstructed for "
                    "file %s" % reconstructionFile
                )
        else:
            global counter_rand
            counter_rand = counter_rand + 1
            return counter_rand

    @docstring(_tsEDFTomoScan.to_dict)
    def to_dict(self):
        res = _tsEDFTomoScan.to_dict(self)
        res.update(TomwerScanBase.to_dict(self))
        return res

    @docstring(TomwerScanBase._deduce_transfert_scan)
    def _deduce_transfert_scan(self, output_dir):
        if os.path.basename(output_dir) != os.path.basename(self.path):
            raise ValueError(
                "Transfert to a new EDFTomoScan requires an "
                "equal basename. Current path is {}, requested "
                "one is {}".format(self.path, output_dir)
            )
        return EDFTomoScan(scan=output_dir)

    def __str__(self):
        return self.path

    # Dataset implementation

    @staticmethod
    def from_dataset_identifier(identifier):
        """Return the Dataset from a identifier"""
        if not isinstance(identifier, _EDFTomoScanIdentifier):
            raise TypeError(
                "identifier should be an instance of " "_HDF5TomoScanIdentifier"
            )
        return EDFTomoScan(scan=identifier.edf_folder)

    @docstring(TomwerScanBase)
    def get_dataset_identifier(self) -> DatasetIdentifier:
        return _EDFTomoScanIdentifier(dataset=self, edf_folder=self.path)

    @docstring(TomwerScanBase)
    def get_nabu_dataset_info(self, binning=1, binning_z=1, proj_subsampling=1):
        return {
            "hdf5_entry": "",
            "location": self.path,
            "binning": binning,
            "binning_z": binning_z,
            "projections_subsampling": proj_subsampling,
        }

    @docstring(TomwerScanBase)
    def to_nabu_dataset_analyser(self):
        from nabu.resources.dataset_analyzer import EDFDatasetAnalyzer

        analyzer = EDFDatasetAnalyzer(location=self.path)
        n_angles = analyzer.n_angles
        if analyzer.rotation_angles is None:
            if self.field_of_view is FOV.HALF:
                angles = numpy.linspace(0, 2 * numpy.pi, n_angles, True)
            else:
                angles = numpy.linspace(0, numpy.pi, n_angles, False)

        analyzer.rotation_angles = angles
        return analyzer

    @docstring(TomwerScanBase)
    def scan_dir_name(self) -> Optional[str]:
        """for 'this/is/my/acquisition' returns 'acquisition'"""
        if self.path is not None:
            return self.path.split(os.sep)[-1]
        else:
            return None

    @docstring(TomwerScanBase)
    def scan_basename(self) -> Optional[str]:
        """for 'this/is/my/acquisition' returns 'acquisition'"""
        if self.path is not None:
            return self.path
        else:
            return None

    @docstring(TomwerScanBase)
    def scan_parent_dir_basename(self) -> Optional[str]:
        if self.path is not None:
            try:
                return os.path.dirname(self.path)
            except Exception:
                return None
        else:
            return None


def _get_urls_list(radio):
    """create the list of urls contained in a .edf file"""
    with fabio.open(radio) as edf_reader:
        if edf_reader.nframes == 1:
            return [DataUrl(file_path=radio, scheme="fabio")]
        else:
            res = []
            for iframe in range(edf_reader.nframes):
                res.append(
                    DataUrl(file_path=radio, data_slice=[iframe], scheme="fabio")
                )
            return res



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "24/01/2017"

import unittest
import logging
import tempfile
import numpy
import fabio.edfimage
import contextlib
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.process.task import Task
from tomwer.test.utils import UtilsTest
from tomwer.core.utils.scanutils import MockEDF
from tomwer.core.process.reconstruction.nabu import nabuslices
import shutil
import os
from multiprocessing import Process

logging.disable(logging.INFO)


class TestScanFactory(unittest.TestCase):
    """Make sure the Scan factory is correctly working. Able to detect the valid
    scan type for a given file / directory
    """

    def test_no_scan(self):
        scan_dir = tempfile.mkdtemp()
        with self.assertRaises(ValueError):
            ScanFactory.create_scan_object(scan_dir)

    def test_scan_edf(self):
        scan_dir = UtilsTest.getEDFDataset("test10")
        scan = ScanFactory.create_scan_object(scan_dir)
        self.assertTrue(isinstance(scan, EDFTomoScan))


class FTSerieReconstructionTest(unittest.TestCase):
    """unit test for the FTSerieReconstruction functions"""

    def setUp(self):
        def saveData(data, _file, folder):
            file_desc = fabio.edfimage.EdfImage(data=data)
            file_desc.write(os.path.join(folder, _file))

        unittest.TestCase.setUp(self)
        self.folderTwoFlats = tempfile.mkdtemp()
        self.folderOneFlat = tempfile.mkdtemp()

        self.darkData = numpy.arange(100).reshape(10, 10)
        self.flatField_0 = numpy.arange(start=10, stop=110).reshape(10, 10)
        self.flatField_600 = numpy.arange(start=-10, stop=90).reshape(10, 10)
        self.flatField = numpy.arange(start=0, stop=100).reshape(10, 10)

        # save configuration one
        saveData(self.darkData, "darkHST.edf", self.folderTwoFlats)
        saveData(self.flatField_0, "refHST_0000.edf", self.folderTwoFlats)
        saveData(self.flatField_600, "refHST_0600.edf", self.folderTwoFlats)

        # save configuration two
        saveData(self.darkData, "darkHST.edf", self.folderOneFlat)
        saveData(self.flatField, "refHST.edf", self.folderOneFlat)

        self.acquiOneFlat = EDFTomoScan(self.folderOneFlat)
        self.acquiTwoFlats = EDFTomoScan(self.folderTwoFlats)

    def tearDown(self):
        for f in (self.folderOneFlat, self.folderTwoFlats):
            shutil.rmtree(f)
        unittest.TestCase.tearDown(self)

    def testGetDark(self):
        self.assertTrue(numpy.array_equal(self.acquiOneFlat.getDark(), self.darkData))
        self.assertTrue(numpy.array_equal(self.acquiTwoFlats.getDark(), self.darkData))

    def testGetFlat(self):
        # check one flat file with two ref
        self.assertTrue(numpy.array_equal(self.acquiOneFlat.getFlat(), self.flatField))

        # check two flat files
        self.assertTrue(
            numpy.array_equal(
                self.acquiTwoFlats.getFlat(), numpy.arange(100).reshape(10, 10)
            )
        )
        self.assertTrue(
            numpy.array_equal(
                self.acquiTwoFlats.getFlat(index=300), numpy.arange(100).reshape(10, 10)
            )
        )
        self.assertTrue(
            numpy.array_equal(self.acquiTwoFlats.getFlat(index=600), self.flatField_600)
        )
        self.assertTrue(
            numpy.array_equal(self.acquiTwoFlats.getFlat(index=0), self.flatField_0)
        )


class TestScanValidatorFindNabuFiles(unittest.TestCase):
    """
    Make sure files produced by nabu can be seen as reconstruction files
    """

    class FileCreator(contextlib.AbstractContextManager):
        def __init__(self, file_path):
            self.__file_path = file_path

        def __enter__(self):
            with open(self.__file_path, "w"):
                pass

        def __exit__(self, exc_type, exc_val, exc_tb):
            if os.path.exists(self.__file_path):
                os.remove(self.__file_path)

    def setUp(self) -> None:
        self.scan = MockEDF.mockScan(tempfile.mkdtemp(), nPagRecons=0, nRecons=0)
        assert (
            len(self.scan.get_reconstructions_paths(self.scan.path)) == 0
        ), "make sure the initial environment is clear"

    def tearDown(self) -> None:
        shutil.rmtree(self.scan.path)

    def test(self) -> None:
        for slice_index in (None, 1, 15):
            for pag_db in (None, 100, 500):
                file_name = (
                    nabuslices.SingleSliceRunner.get_file_basename_reconstruction(
                        slice_index=slice_index,
                        scan=self.scan,
                        pag=pag_db is not None,
                        db=pag_db,
                    )
                )
                for file_extension in ("npy", "npz", "hdf5"):
                    with self.subTest(
                        slice_index=slice_index,
                        pag_db=pag_db,
                        file_extension=file_extension,
                    ):
                        file_path = os.path.join(
                            self.scan.path, file_name + "." + file_extension
                        )

                        # simple context Manager which create the file
                        # and remove it when leave
                        with TestScanValidatorFindNabuFiles.FileCreator(file_path):
                            n_to_discover = 1
                            if slice_index is None:
                                n_to_discover = 0
                            self.assertEqual(
                                len(
                                    self.scan.get_reconstructions_paths(self.scan.path)
                                ),
                                n_to_discover,
                            )


class TestScanValidatorFindPyHSTFiles(unittest.TestCase):
    """Function testing the getReconstructionsPaths function is correctly
    functioning"""

    DIM_MOCK_SCAN = 10

    N_RADIO = 20
    N_RECONS = 10
    N_PAG_RECONS = 5

    def setUp(self):
        # create scan folder
        self.path = tempfile.mkdtemp()
        MockEDF.mockScan(
            scanID=self.path,
            nRadio=self.N_RADIO,
            nRecons=self.N_RECONS,
            nPagRecons=self.N_PAG_RECONS,
            dim=self.DIM_MOCK_SCAN,
        )
        basename = os.path.basename(self.path)

        # add some random files
        for _file in ("45gfdgfg1.edf", "465slicetest1.edf", "slice_ab.edf"):
            with open(os.path.join(self.path, basename + _file), "w+") as ofile:
                ofile.write("test")

    def tearDown(self):
        if os.path.isdir(self.path):
            shutil.rmtree(self.path)

    def testGetRadioPaths(self):
        nFound = len(EDFTomoScan.get_proj_urls(self.path))
        self.assertTrue(nFound == self.N_RADIO)

    def testGetReconstructionsPaths(self):
        reconstruction_found = EDFTomoScan.get_reconstructions_paths(self.path)
        self.assertEqual(len(reconstruction_found), self.N_RECONS + self.N_PAG_RECONS)

    def testGetSliceReconstruction(self):
        self.assertTrue(
            EDFTomoScan.get_index_reconstructed("dfadf_slice_32.edf", "dfadf") == 32
        )
        self.assertTrue(
            EDFTomoScan.get_index_reconstructed("dfadf_slice_slice_002.edf", "dfadf")
            == 2
        )
        self.assertTrue(
            EDFTomoScan.get_index_reconstructed("scan_3slice_0050.edf", "scan_3") == 50
        )
        self.assertTrue(
            EDFTomoScan.get_index_reconstructed("scan3slice_012.edf", "scan3") == 12
        )


class TestRadioPath(unittest.TestCase):
    """Test static method getRadioPaths for EDFTomoScan"""

    def test(self):
        files = [
            "essai1_0008.edf",
            "essai1_0019.edf",
            "essai1_0030.edf",
            "essai1_0041.edf",
            "essai1_0052.edf",
            "essai1_0063.edf",
            "essai1_0074.edf",
            "essai1_0085.edf",
            "essai1_0096.edf",
            "essai1_.par",
            "refHST0100.edf",
            "darkend0000.edf",
            "essai1_0009.edf",
            "essai1_0020.edf",
            "essai1_0031.edf",
            "essai1_0042.edf",
            "essai1_0053.edf",
            "essai1_0064.edf",
            "essai1_0075.edf",
            "essai1_0086.edf",
            "essai1_0097.edf",
            "essai1_.rec",
            "essai1_0000.edf",
            "essai1_0010.edf",
            "essai1_0021.edf",
            "essai1_0032.edf",
            "essai1_0043.edf",
            "essai1_0054.edf",
            "essai1_0065.edf",
            "essai1_0076.edf",
            "essai1_0087.edf",
            "essai1_0098.edf",
            "essai1_slice_1023.edf",
            "essai1_0001.edf",
            "essai1_0011.edf",
            "essai1_0022.edf",
            "essai1_0033.edf",
            "essai1_0044.edf",
            "essai1_0055.edf",
            "essai1_0066.edf",
            "essai1_0077.edf",
            "essai1_0088.edf",
            "essai1_0099.edf",
            "essai1_slice.info",
            "essai1_0001.par",
            "essai1_0012.edf",
            "essai1_0023.edf",
            "essai1_0034.edf",
            "essai1_0045.edf",
            "essai1_0056.edf",
            "essai1_0067.edf",
            "essai1_0078.edf",
            "essai1_0089.edf",
            "essai1_0100.edf",
            "essai1_slice.par",
            "essai1_0002.edf",
            "essai1_0013.edf",
            "essai1_0024.edf",
            "essai1_0035.edf",
            "essai1_0046.edf",
            "essai1_0057.edf",
            "essai1_0068.edf",
            "essai1_0079.edf",
            "essai1_0090.edf",
            "essai1_0101.edf",
            "essai1_slice.xml",
            "essai1_0003.edf",
            "essai1_0014.edf",
            "essai1_0025.edf",
            "essai1_0036.edf",
            "essai1_0047.edf",
            "essai1_0058.edf",
            "essai1_0069.edf",
            "essai1_0080.edf",
            "essai1_0091.edf",
            "essai1_0102.edf",
            "essai1_.xml",
            "essai1_0004.edf",
            "essai1_0015.edf",
            "essai1_0026.edf",
            "essai1_0037.edf",
            "essai1_0048.edf",
            "essai1_0059.edf",
            "essai1_0070.edf",
            "essai1_0081.edf",
            "essai1_0092.edf",
            "essai1_0103.edf",
            "histogram_essai1_slice",
            "essai1_0005.edf",
            "essai1_0016.edf",
            "essai1_0027.edf",
            "essai1_0038.edf",
            "essai1_0049.edf",
            "essai1_0060.edf",
            "essai1_0071.edf",
            "essai1_0082.edf",
            "essai1_0093.edf",
            "essai1_0104.edf",
            "machinefile",
            "essai1_0006.edf",
            "essai1_0017.edf",
            "essai1_0028.edf",
            "essai1_0039.edf",
            "essai1_0050.edf",
            "essai1_0061.edf",
            "essai1_0072.edf",
            "essai1_0083.edf",
            "essai1_0094.edf",
            "essai1_.cfg",
            "pyhst_out.txt",
            "essai1_0007.edf",
            "essai1_0018.edf",
            "essai1_0029.edf",
            "essai1_0040.edf",
            "essai1_0051.edf",
            "essai1_0062.edf",
            "essai1_0073.edf",
            "essai1_0084.edf",
            "essai1_0095.edf",
        ]

        nbRadio = 0
        for f in files:
            nbRadio += EDFTomoScan.is_a_proj_path(f, "essai1_")
        self.assertTrue(nbRadio == 105)


class TestGetReconstructionPath(unittest.TestCase):
    """Test static method getReconstructionPaths for EDFTomoScan"""

    def setUp(self):
        self.tmpDir = tempfile.mkdtemp()
        self.dir = os.path.join(self.tmpDir, "scan_3")
        os.makedirs(self.dir)

    def tearDown(self):
        shutil.rmtree(self.tmpDir)

    def test1(self):
        """Make sure he can read 3 simple slice"""
        file_desc = fabio.edfimage.EdfImage(data=numpy.random.random((100, 100)))

        file_desc.write(os.path.join(self.dir, "scan_3slice_0050.edf"))
        file_desc.write(os.path.join(self.dir, "scan_3slice_0000.edf"))
        file_desc.write(os.path.join(self.dir, "scan_3slice_0010.edf"))
        assert len(os.listdir(self.dir)) == 3
        self.assertEqual(
            len(
                EDFTomoScan.get_reconstructions_paths(
                    os.path.join(self.tmpDir, "scan_3")
                )
            ),
            3,
        )

    def test2(self):
        """Make sure he can read a paganin and non-paganin at the same time"""
        file_desc = fabio.edfimage.EdfImage(data=numpy.random.random((100, 100)))

        file_desc.write(os.path.join(self.dir, "scan_3slice_0050.edf"))
        file_desc.write(os.path.join(self.dir, "scan_3slice_pag_0050.edf"))
        assert len(os.listdir(self.dir)) == 2
        self.assertEqual(
            len(
                EDFTomoScan.get_reconstructions_paths(
                    os.path.join(self.tmpDir, "scan_3")
                )
            ),
            2,
        )


class TestPaganinPath(unittest.TestCase):
    """Test static method getIndexReconstructed from EDFTomoScan"""

    def setUp(self):
        self.tmpDir = tempfile.mkdtemp()

        self.dir = os.path.join(self.tmpDir, "scan25")
        os.makedirs(self.dir)

    def tearDown(self):
        shutil.rmtree(self.tmpDir)

    def testPagFile(self):
        open(os.path.join(self.dir, "scan25slice_pag_db0500_0115.edf"), "a").close()
        self.assertEqual(len(EDFTomoScan.get_reconstructions_paths(self.dir)), 1)
        self.assertEqual(
            EDFTomoScan.get_index_reconstructed(
                "scan25slice_pag_db0500_0115.edf", scanID=self.dir
            ),
            115,
        )


class TestGetSinogram(unittest.TestCase):
    """Test functions relative to sinogram"""

    def setUp(self) -> None:
        self.tmp_dir = tempfile.mkdtemp()
        mock = MockEDF(
            scan_path=self.tmp_dir, scene="arange", n_radio=4, n_ini_radio=4, dim=2
        )
        mock.end_acquisition()
        self.scan = EDFTomoScan(scan=mock.scan_path)
        self.assertEqual(len(self.scan.projections), 4)

    def tearDown(self) -> None:
        shutil.rmtree(self.tmp_dir)

    def testGetSinogram(self):
        self.assertEqual(self.scan.tomo_n, 4)
        sinogram_1 = self.scan.get_sinogram(line=1)
        self.assertEqual(sinogram_1.shape, (4, 2))
        th_res = numpy.array([[2, 3], [6, 7], [10, 11], [14, 15]])
        self.assertTrue(numpy.array_equal(sinogram_1, th_res))


class TestSeveralWriteWithThreading(unittest.TestCase):
    class DummyProcess(Task):
        @staticmethod
        def program_name():
            """Name of the program used for this processing"""
            return "dummy program"

        @staticmethod
        def program_version():
            """version of the program used for this processing"""
            return "0.0.0"

        @staticmethod
        def definition():
            """definition of the process"""
            return "no definition"

    def setUp(self):
        self.tmp_dir = tempfile.mkdtemp()
        self._file = os.path.join(self.tmp_dir, "XXXtomwer_processes.h5")

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)

    def testSeveralWrite(self):
        processes = []
        for i in range(20):
            args = (
                self._file,
                TestSeveralWriteWithThreading.DummyProcess,
                "entry0000",
                {"info": "no configuration"},
                {"data": numpy.random.random(2000)},
                i,
                None,
                True,
            )
            processes.append(Process(target=Task._register_process, args=args))
        for process in processes:
            process.start()
        for process in processes:
            process.join()



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "16/11/2020"


import logging
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.process.task import Task
from tomwer.core.scan.scanfactory import ScanFactory
from silx.utils.enum import Enum as _Enum
from tomwer.utils import docstring
import tomwer.version
import os

_logger = logging.getLogger(__name__)


class OutputType(_Enum):
    STATIC = "static"
    ONE_LEVEL_UPPER = "../volume"


def create_volume_symbolic_link(scan: TomwerScanBase, output_folder: str):
    """
    Create a symbolic link for each volume reconstructed of `scan`

    :param TomwerScanBase scan:
    :param str output_folder:
    """
    if scan.latest_vol_reconstructions is None:
        _logger.info("No volume reconstructed found")
        return
    if not os.path.exists(output_folder):
        os.makedirs(output_folder)
    for volume_url in scan.latest_vol_reconstructions:
        dst = os.path.join(output_folder, os.path.basename(volume_url.file_path()))
        if os.path.exists(dst):
            _logger.info("{} already exists. Cannot create a symbolic link on it")
        else:
            os.symlink(src=volume_url.file_path(), dst=dst)


class VolumeSymbolicLinkProcess(Task, input_names=("data",), output_names=("data",)):
    """
    Process class for volume symbolic link
    """

    def __init__(self, varinfo=None, inputs=None, label=None):
        super().__init__(varinfo=varinfo, inputs=inputs, label=label)
        self._output_type = OutputType.ONE_LEVEL_UPPER
        self._output_folder = None

    @docstring(Task.set_configuration)
    def set_configuration(self, properties):
        if "output_type" in properties:
            self._output_type = OutputType.from_value(properties["output_type"])
        if "output_folder" in properties:
            self._output_folder = properties["output_folder"]

    @docstring(Task.program_name)
    @staticmethod
    def program_name():
        return "tomwer_volume_symlink"

    @docstring(Task.program_version)
    @staticmethod
    def program_version():
        return tomwer.version.version

    @docstring(Task.run)
    def run(self):
        scan = self.inputs.data
        if type(scan) is dict:
            scan = ScanFactory.create_scan_object_frm_dict(scan)
        else:
            scan = scan
        if scan is None:
            return
        if not isinstance(scan, TomwerScanBase):
            raise TypeError(
                "scan is expected to be a dict or an instance "
                "of TomwerScanBase. Not {}".format(type(scan))
            )

        if self._output_type is OutputType.STATIC:
            if self._output_folder is None:
                raise ValueError(
                    "Manual setting of the output folder is "
                    "requested but None is provided."
                )
            else:
                output_folder = self._output_folder
                if not os.path.isabs(output_folder):
                    os.path.abspath(
                        os.path.join(os.path.realpath(scan.path), output_folder)
                    )
        elif self._output_type is OutputType.ONE_LEVEL_UPPER:
            output_folder = os.path.realpath(scan.path)
            output_folder = os.path.abspath(
                os.path.join(output_folder, self._output_type.value)
            )
        else:
            raise ValueError(
                "output type {} is not managed".format(self._output_type.value)
            )
        create_volume_symbolic_link(scan=scan, output_folder=output_folder)
        self.outputs.data = scan

    @docstring(Task.definition)
    @staticmethod
    def definition():
        return "Create a symbolic link to the volume folder"



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import logging
import sys
import argparse
from ewokscore import load_graph
from ewoksorange.bindings import owsconvert
from tomwer.core.scan.scanfactory import ScanFactory
import tempfile
import os
from pprint import pprint

logging.basicConfig(level=logging.WARNING)
_logger = logging.getLogger(__name__)


def exec_scans(name, ewoks_graph, scans=None, darkref_savedir=None):
    """

    :param ewoks_graph:
    :param scans: list of scan to execute
    :param timeout:
    :return:
    """

    def launch_():
        for scan in scans:
            exec_scan(name=name, ewoks_graph=ewoks_graph, scan=scan, save_dir=save_dir)

    if darkref_savedir is not None:
        if not os.path.exists(darkref_savedir):
            os.makedirs(darkref_savedir)
        launch_()
    else:
        with tempfile.TemporaryDirectory() as save_dir:
            launch_()


def exec_scan(name, ewoks_graph, scan, save_dir):
    # set up workflow
    if len(ewoks_graph.start_nodes()) == 0:
        _logger.warning("no start nodes found. Enable to process")
        return None
    mess = f"start processing {name} with {str(scan)}"
    _logger.info(mess)

    # update scan parameter to start node
    for src_node in ewoks_graph.start_nodes():
        # provide the dataset to be used and save_dir
        # save_dir is used to provide dark ref with a directory to store
        # dark and ref to be copy
        ewoks_graph.graph.nodes[src_node]["default_inputs"].add(
            {({"name": "data", "value": scan})}
        )

    # update node with dark ref
    import tomwer.core.process.reconstruction.darkref.darkrefscopy

    qual_name = ".".join(
        [
            tomwer.core.process.reconstruction.darkref.darkrefscopy.__name__,
            tomwer.core.process.reconstruction.darkref.darkrefscopy.DarkRefsCopy.__qualname__,
        ]
    )
    for src_node in ewoks_graph.graph.nodes:
        # provide save_dir to 'darkref' class
        if ewoks_graph.graph.nodes[src_node]["class"] == qual_name:
            ewoks_graph.graph.nodes[src_node]["default_inputs"].add(
                {({"name": "save_dir", "value": save_dir})}
            )
        # insure all arguments are provided for each link because the scan object is hidding some
        # parameters
        for target_node in ewoks_graph.graph.nodes:
            if src_node == target_node:
                continue
            try:
                ewoks_graph.graph[src_node][target_node]["all_arguments"] = True
                del ewoks_graph.graph[src_node][target_node]["arguments"]
            except KeyError:
                pass

    print(f" start execution of graph {name} ".center(80, "#"))
    taskgraph = load_graph(ewoks_graph)
    pprint(taskgraph.dump())
    print(f" end execution of graph {name} ".center(80, "#"))
    ewoks_graph.execute()


def main(argv):
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "workflow_file",
        help="Path to the .ows file defining the workflow to process with the"
        "provided scan",
    )
    parser.add_argument(
        "scan_path",
        help="Path to data to be processes (master file if come from an hdf5 "
        "acquisition or EDF files folder if come from an EDF acquisition)",
    )
    parser.add_argument(
        "--entry", default=None, help="An entry should be specify for hdf5 files"
    )
    parser.add_argument(
        "--timeout", default=None, help="Timeout for the workflow execution"
    )
    parser.add_argument(
        "--dkref-save-dir",
        default=None,
        dest="savedir",
        help="You can provide a save dir folder for the dark-ref copy task to"
        "insure persistency between several call of process.",
    )
    parser.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        default=False,
        help="Set logging system in debug mode",
    )
    options = parser.parse_args(argv[1:])
    if options.entry is not None:
        scan = ScanFactory.create_scan_object(
            options.scan_path, entry=options.entry, accept_bliss_scan=True
        )
        scans = (scan,)
    else:
        scans = ScanFactory.create_scan_objects(
            options.scan_path, accept_bliss_scan=True
        )
    if len(scans) > 1:
        _logger.info("More than one scan found. Will process every scans")

    # tune the log level
    log_level = logging.INFO
    if options.debug is True:
        log_level = logging.DEBUG

    for log_ in ("tomwer", "ewoks", "ewoksorange", "ewokscore"):
        logging.getLogger(log_).setLevel(log_level)

    scheme = owsconvert.ows_to_ewoks(options.workflow_file)
    exec_scans(
        name=os.path.basename(options.workflow_file),
        ewoks_graph=scheme,
        scans=scans,
        darkref_savedir=options.savedir,
    )


if __name__ == "__main__":
    main(sys.argv)



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "04/03/2019"


from collections import OrderedDict
import datetime
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.blissscan import BlissScan
from silx.io.url import DataUrl
from silx.io.utils import h5py_read_dataset
from tomwer.core.process.reconstruction.scores import ComputedScore
import h5py
import typing
import logging

_logger = logging.getLogger(__name__)


def _assert_param_instance(param, valid_instances):
    """check the parameter value, can be an instance of valid_instance or
    an iterable of valid_instance"""
    if isinstance(param, valid_instances):
        return
    else:
        try:
            for unique_param in param:
                if not isinstance(unique_param, valid_instances):
                    assert isinstance(unique_param, valid_instances)
        except Exception:
            raise ValueError("%s has an invalid type" % param)


def _assert_cast_to_boolean(value):
    """

    :param value: value we want to cast to a boolean
    """
    try:
        bool(int(value))
    except Exception:
        raise ValueError("%s cannot be cast to boolean" % value)


def _assert_cast_to_int(value):
    try:
        int(value)
    except Exception:
        raise ValueError("%s cannot be cast to int" % value)


class LastReceivedScansDict(OrderedDict):
    """List the received scan from the first received to the last received"""

    def __init__(self, limit=None):
        """Simple structure in order to store the received last elements and
        the time of acquisition
        """
        assert limit is None or (type(limit) is int and limit > 0)
        OrderedDict.__init__(self)
        self.limit = limit

    def add(self, scan):
        assert isinstance(scan, (TomwerScanBase, BlissScan))
        self[str(scan)] = datetime.datetime.now()
        if self.limit is not None and len(self) > self.limit:
            self.pop(list(self.keys())[0])


class IgnoreProcess:
    """Simple util class to ignore a processing when using ewoks"""

    def process(self, scan):
        return scan

    __call__ = process


def get_scores(node: h5py.Group) -> typing.Union[None, dict]:
    """read all the score from the 'results' sub group. Read the url
    and computed scores

    :return: a dictionary with (url, ComputedScore) for each node
    :rtype: dict or None
    """
    if "results" not in node:
        _logger.warning(f"no results found in {node}")
        return None
    else:
        res_node = node["results"]
        scores = {}
        for res_item in res_node:
            score_node = res_node[res_item]
            try:
                float(res_item)
            except Exception:
                is_a_score = False
            else:
                is_a_score = True

            if not is_a_score:
                continue

            try:
                url = DataUrl(
                    file_path=node.file.filename,
                    data_path="/".join((score_node.name, "reconstructed_slice")),
                    scheme="silx",
                )
                score = ComputedScore(
                    tv=h5py_read_dataset(score_node["total variation"]),
                    std=h5py_read_dataset(score_node["standard deviation"]),
                )
                scores[float(res_item)] = (url, score)
            except Exception as e:
                _logger.warning(
                    "Unable to load {}. Reason is {}".format(score_node.name, str(e))
                )
        scores = dict(sorted(scores.items(), key=lambda item: item[0]))

        return scores



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2016-2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "14/10/2021"


from orangewidget import gui
from orangewidget.widget import Input, Output
from orangewidget.widget import OWBaseWidget
from orangewidget.settings import Setting
from silx.gui import qt
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.futurescan import FutureTomwerScan
from tomwer.gui.cluster.supervisor import (
    FutureTomwerScanObserverWidget as _FutureTomwerScanObserverWidget,
)
from processview.core.superviseprocess import SuperviseProcess
from processview.core.manager import ProcessManager, DatasetState
from silx.gui.utils.concurrent import submitToQtMainThread
import time


class FutureSupervisorOW(OWBaseWidget, openclass=True):
    """
    Orange widget to define a slurm cluster as input of other
    widgets (based on nabu for now)
    """

    name = "future supervisor"
    id = "orange.widgets.tomwer.cluster.FutureSupervisorOW.FutureSupervisorOW"
    description = "Observe slurm job registered."
    icon = "icons/slurmobserver.svg"
    priority = 22
    keywords = [
        "tomography",
        "tomwer",
        "slurm",
        "observer",
        "cluster",
        "job",
        "sbatch",
        "supervisor",
        "future",
    ]

    want_main_area = True
    resizing_enabled = True
    compress_signal = False

    _static_input = Setting(dict())

    class Inputs:
        future_in = Input(
            name="future_data",
            type=FutureTomwerScan,
            doc="data with some remote processing",
            multiple=True,
            default=True,
        )

    class Outputs:
        data = Output(name="data", type=TomwerScanBase)

    def __init__(self, parent=None):
        super().__init__(parent)
        # gui
        layout = gui.vBox(self.mainArea, self.name).layout()
        self._widget = FutureTomwerScanObserverWidget(
            parent=self, name=self.windowTitle()
        )
        layout.addWidget(self._widget)

        # connect signal / slot
        self._widget.observationTable.model().sigStatusUpdated.connect(
            self._convertBackAutomatically
        )
        self._widget.sigConversionRequested.connect(self._convertBack)

    def convertWhenFinished(self):
        return self._widget.convertWhenFinished()

    def _convertBackAutomatically(self, future_scan, status):
        if not isinstance(future_scan, FutureTomwerScan):
            raise TypeError(
                f"future_scan is expected to be an instance of {FutureTomwerScan} and not {type(future_scan)}"
            )
        if status == "finished" and self.convertWhenFinished():
            self._convertBack(future_scan)

    def _convertBack(self, future_scan):
        if not isinstance(future_scan, FutureTomwerScan):
            raise TypeError(
                f"future_scan is expected to be an instance of {FutureTomwerScan} and not {type(future_scan)}"
            )
        self._futureHasBeenConverted(future_scan, future_scan.scan)

    @Inputs.future_in
    def add(self, future_scan, signal_id=None):
        if future_scan is not None:
            self._widget.addFutureScan(future_scan=future_scan)

    def _futureHasBeenConverted(self, future_scan, scan):
        # clean client to free resources
        self._widget.removeFutureScan(future_scan=future_scan)
        future_scan.clear_clients()
        if scan is not None:
            self.Outputs.data.send(scan)


class FutureTomwerScanObserverWidget(_FutureTomwerScanObserverWidget, SuperviseProcess):
    """add dataset state notification (ProcessManager) to the original FutureTomwerScanObserverWidget"""

    REFRESH_FREQUENCE = 2
    """time between call to updateView"""

    def __init__(self, name, parent=None):
        super().__init__(parent=parent)
        self.name = name
        self._updateThread = _RefreshThread(
            callback=self.updateView, refresh_frequence=self.REFRESH_FREQUENCE
        )
        self.destroyed.connect(self.stopRefresh)
        self._updateThread.start()

    def stopRefresh(self):
        if self._updateThread is not None:
            self._updateThread.stop()
            self._updateThread.wait(self.REFRESH_FREQUENCE + 1)
            self._updateThread = None

    def close(self):
        self.stopRefresh()
        super().close()

    def addFutureScan(self, future_scan: FutureTomwerScan):
        super().addFutureScan(future_scan)
        self._updateScanSupervisor(future_scan)

    def removeFutureScan(self, future_scan: FutureTomwerScan):
        self._updateScanSupervisor(future_scan)
        super().removeFutureScan(future_scan)

    def _updateScanSupervisor(self, future_scan):
        r_id = future_scan.process_requester_id
        if r_id is not None:
            requester_name = ProcessManager().get_process(r_id).name
        else:
            requester_name = "unknow"
        details = f"job spawn by {requester_name}"
        if future_scan is None:
            return
        elif future_scan.status == "error":
            state = DatasetState.FAILED
            details = "\n".join([details, future_scan.exception()])
        elif future_scan.status == "pending":
            details = "\n".join([details, "pending"])
            state = DatasetState.PENDING
        elif future_scan.status == "finished":
            state = DatasetState.SUCCEED
        elif future_scan.status == "running":
            details = "\n".join([details, "running"])
            state = DatasetState.ON_GOING
        elif future_scan.status == "cancelled":
            details = "\n".join([details, "job cancelled"])
            state = DatasetState.SKIPPED
        elif future_scan.status is None:
            return
        else:
            raise ValueError(f"future scan status {future_scan.status} is not managed")
        ProcessManager().notify_dataset_state(
            dataset=future_scan.scan,
            process=self,
            state=state,
            details=details,
        )

    def _updateStatus(self, future_scan):
        self._updateScanSupervisor(future_scan)
        super()._updateStatus(future_scan)


class _RefreshThread(qt.QThread):
    """Simple thread to call a refresh callback each refresh_frequence (seconds)"""

    TIME_BETWEEN_LOOP = 0.5

    def __init__(self, callback, refresh_frequence) -> None:
        super().__init__()
        self._callback = callback
        self._refresh_frequence = refresh_frequence
        self._stop = False

    def stop(self):
        self._stop = True
        self._callback = None

    def run(self):
        w_t = self._refresh_frequence + self.TIME_BETWEEN_LOOP

        while not self._stop:
            if w_t <= 0:
                if self._callback is not None:
                    try:
                        submitToQtMainThread(self._callback)
                    except AttributeError:
                        # can happen when closing
                        pass
                w_t = self._refresh_frequence + self.TIME_BETWEEN_LOOP
            w_t -= self.TIME_BETWEEN_LOOP
            time.sleep(self.TIME_BETWEEN_LOOP)



# coding: utf-8
# /*##########################################################################
# Copyright (C) 2016 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "14/02/2020"


import os
import shutil
import tempfile
from glob import glob
import unittest
from tomwer.test.utils import UtilsTest
from tomwer.core.process.reconstruction.nabu import nabuslices as nabu
from tomwer.core.process.reconstruction.nabu import utils as nabuutils
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.process.reconstruction.nabu.nabuslices import NabuSliceMode
from tomwer.core.utils.scanutils import MockEDF
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.process.reconstruction.axis import AxisRP
from tomwer.core.process.reconstruction.nabu import settings as nabu_settings


class TestNabuDescUtils(unittest.TestCase):
    """Test the util function for creating the description of nabu"""

    def setUp(self):
        unittest.TestCase.setUp(self)
        self.top_dir = tempfile.mkdtemp()
        self.dataset = "scan_3_"
        data_test_dir = UtilsTest.getEDFDataset(self.dataset)
        scan_folder = os.path.join(self.top_dir, self.dataset)
        shutil.copytree(data_test_dir, scan_folder)
        self.scan = ScanFactory.create_scan_object(scan_folder)

    def tearDown(self) -> None:
        shutil.rmtree(self.scan.path)

    def testResourcesDescriptor(self):
        nabuutils.get_nabu_resources_desc(scan=self.scan, method="local")
        nabuutils.get_nabu_resources_desc(scan=self.scan, method="slurm")

    def testDatasetDescriptor(self):
        self.scan.get_nabu_dataset_info()


class TestNabuIni(unittest.TestCase):
    """Test creation of nabu .cfg files according to tomwer configuration"""

    def setUp(self) -> None:
        self.dataset = "scan_3_"
        self._root_dir = tempfile.mkdtemp()
        data_test_dir = UtilsTest.getEDFDataset(self.dataset)
        scan_folder = os.path.join(self._root_dir, self.dataset)
        shutil.copytree(data_test_dir, scan_folder)
        self.scan = EDFTomoScan(scan_folder)
        self._nabu_cfg_folder = os.path.join(
            self.scan.path, nabu_settings.NABU_CFG_FILE_FOLDER
        )
        [os.remove(cfg_file) for cfg_file in glob(self.scan.path + os.sep + "*.cfg")]
        if os.path.exists(self._nabu_cfg_folder):
            shutil.rmtree(self._nabu_cfg_folder)

        assert len(self.getNabuIniFiles()) == 0
        # a random configuration
        self.config = get_nabu_default_config(output_dir=self.scan.path)

    def tearDown(self) -> None:
        shutil.rmtree(self._root_dir)

    def testNoSlices(self):
        """If no slices selected, in this case we will only create a .cfg
        for the volume
        """
        self.config["tomwer_slices"] = None
        self.config["phase"]["method"] = ""
        self.assertEqual(
            len(
                nabu.interpret_tomwer_configuration(config=self.config, scan=self.scan)
            ),
            1,
        )
        nabu.run_slices_reconstruction(scan=self.scan, config=self.config, dry_run=True)
        nabu_files = self.getNabuIniFiles()
        self.assertEqual(len(nabu_files), 1)
        # check volume and slice .cfg files for nabu reconstruction are here
        ini_file = os.path.basename(self.scan.path) + ".cfg"
        ini_file = os.path.join(self._nabu_cfg_folder, ini_file)
        self.assertTrue(ini_file in nabu_files)

    def testSingleSlice(self):
        """Test the behavior if we only request one slice without Paganin"""
        self.config["tomwer_slices"] = "1"
        self.config["phase"]["method"] = ""
        self.assertEqual(
            len(
                nabu.interpret_tomwer_configuration(config=self.config, scan=self.scan)
            ),
            2,
        )
        nabu.run_slices_reconstruction(scan=self.scan, config=self.config, dry_run=True)
        nabu_files = self.getNabuIniFiles()
        self.assertEqual(len(nabu_files), 2)
        # check volume and slice .cfg files for nabu reconstruction are here
        for file_name in (
            os.path.basename(self.scan.path) + ".cfg",
            os.path.basename(self.scan.path) + "slice_0001" + ".cfg",
        ):
            with self.subTest(file_name=file_name):
                self.assertTrue(
                    os.path.join(self._nabu_cfg_folder, file_name) in nabu_files
                )

    def testSinglePagSlice(self):
        """Test the behavior if we only request one Paganin slice"""
        self.config["phase"]["method"] = "Paganin"
        self.config["tomwer_slices"] = "1"
        self.config["phase"]["delta_beta"] = "100.0"
        self.assertEqual(
            len(
                nabu.interpret_tomwer_configuration(config=self.config, scan=self.scan)
            ),
            2,
        )
        nabu.run_slices_reconstruction(scan=self.scan, config=self.config, dry_run=True)
        nabu_files = self.getNabuIniFiles()
        self.assertEqual(len(nabu_files), 2)
        # check volume and slice .cfg files for nabu reconstruction are here
        for file_name in (
            os.path.basename(self.scan.path) + "pag_db0100.cfg",
            os.path.basename(self.scan.path) + "slice_pag_0001_db0100" + ".cfg",
        ):
            with self.subTest(file_name=file_name):
                self.assertTrue(
                    os.path.join(self._nabu_cfg_folder, file_name) in nabu_files
                )

    def testSeveralSlices(self):
        """Test the behavior if we request several slices without Paganin"""
        slices = (1, 4, 12)
        self.config["tomwer_slices"] = str(slices)
        self.config["phase"]["method"] = ""
        self.assertEqual(
            len(
                nabu.interpret_tomwer_configuration(config=self.config, scan=self.scan)
            ),
            len(slices) + 1,
        )
        nabu.run_slices_reconstruction(scan=self.scan, config=self.config, dry_run=True)
        nabu_files = self.getNabuIniFiles()
        self.assertEqual(len(nabu_files), len(slices) + 1)
        # check volume and slice .ini files for nabu reconstruction are here
        ini_files = [os.path.basename(self.scan.path) + ".cfg"]
        for slice_index in slices:
            ini_files.append(
                os.path.basename(self.scan.path)
                + "slice_"
                + str(slice_index).zfill(4)
                + ".cfg"
            )

        for ini_file in ini_files:
            with self.subTest(file_name=ini_file):
                self.assertTrue(
                    os.path.join(self._nabu_cfg_folder, ini_file) in nabu_files
                )

    def testSeveralPags(self):
        """behavior when several paganin slices are requested"""
        self.config["tomwer_slices"] = "1"
        self.config["phase"]["method"] = "Paganin"
        pag_dbs = [200.0, 300.0, 600.0, 700.0]
        self.config["phase"]["delta_beta"] = str(pag_dbs)
        sub_configs = nabu.interpret_tomwer_configuration(
            config=self.config, scan=self.scan
        )
        self.assertEqual(len(sub_configs), len(pag_dbs) * 2)
        # check that each dbs and slices values are in sub configurations
        existing_dbs = [config[0]["phase"]["delta_beta"] for config in sub_configs]
        existing_dbs = set(existing_dbs)
        for pag_db in pag_dbs:
            with self.subTest(pag_db=pag_db):
                self.assertTrue(str(pag_db) in existing_dbs)

        nabu.run_slices_reconstruction(scan=self.scan, config=self.config, dry_run=True)
        nabu_files = self.getNabuIniFiles()
        self.assertEqual(len(nabu_files), len(pag_dbs) * 2)

        # check volume and slice .conf files for nabu reconstruction are here
        ini_files = []
        for pag_db in pag_dbs:
            # add volume ini file
            ini_files.append(
                os.path.basename(self.scan.path)
                + "pag_db"
                + str(int(pag_db)).zfill(4)
                + ".cfg"
            )
            # add slice ini file
            ini_files.append(
                os.path.basename(self.scan.path)
                + "slice_pag_0001_db"
                + str(int(pag_db)).zfill(4)
                + ".cfg"
            )

        for ini_file in ini_files:
            with self.subTest(file_name=ini_file):
                self.assertTrue(
                    os.path.join(self._nabu_cfg_folder, ini_file) in nabu_files
                )

    def testSeveralPagsSeveralSlices(self):
        """behavior when several paganin slices are requested with several
        delta / beta values"""
        slices = (1, 4, 12)
        self.config["tomwer_slices"] = str(slices)
        self.config["phase"]["method"] = "Paganin"
        pag_dbs = [200, 300, 600, 700]
        self.config["phase"]["delta_beta"] = str(pag_dbs)
        self.assertEqual(
            len(
                nabu.interpret_tomwer_configuration(config=self.config, scan=self.scan)
            ),
            len(slices) * len(pag_dbs) + len(pag_dbs),
        )
        nabu.run_slices_reconstruction(scan=self.scan, config=self.config, dry_run=True)
        nabu_files = self.getNabuIniFiles()
        self.assertEqual(len(nabu_files), len(slices) * len(pag_dbs) + len(pag_dbs))
        # check volume and slice .cfg files for nabu reconstruction are here
        ini_files = []
        for slice_index in slices:
            for pag_db in pag_dbs:
                # add volume ini file
                ini_files.append(
                    os.path.basename(self.scan.path)
                    + "pag_db"
                    + str(pag_db).zfill(4)
                    + ".cfg"
                )
                # add slice ini file
                ini_files.append(
                    os.path.basename(self.scan.path)
                    + "slice_pag_"
                    + str(slice_index).zfill(4)
                    + "_db"
                    + str(pag_db).zfill(4)
                    + ".cfg"
                )

        for ini_file in ini_files:
            with self.subTest(file_name=ini_file):
                self.assertTrue(
                    os.path.join(self._nabu_cfg_folder, ini_file) in nabu_files
                )

    def getNabuIniFiles(self):
        return glob(self._nabu_cfg_folder + os.sep + "*.cfg")


class TestTomwerConfForNabu(unittest.TestCase):
    """
    Test the nabu configuration created from tomwer.
    Tomwer can request several slice or several paganin db values.
    We insure we can retrieve the correct nabu configuration for those

    note: the `_interpret_tomwer_configuration` function create each time a
          'None' slice which is the reconstruction of the entire volume
    """

    def setUp(self) -> None:
        self.scan = MockEDF.mockScan(scanID=tempfile.mkdtemp())
        # initial configuration
        self.conf = {"reconstruction": {}}

    def tearDown(self) -> None:
        shutil.rmtree(self.scan.path)

    def testSlices(self):
        # check if single value
        for slice in ("1", NabuSliceMode.MIDDLE):
            with self.subTest(slice=slice):
                self.conf["tomwer_slices"] = slice

                sub_confs = nabu.interpret_tomwer_configuration(
                    self.conf, scan=self.scan
                )
                # not 1 because return each time the 'None' slice which is
                # the reconstruction of the entire volume
                self.assertEqual(len(sub_confs), 1 + 1)

        # check single values or list of values
        for slice in ("[2, 3, 7]", "(6, 9,56)", "5;6; 9"):
            with self.subTest(slice=slice):
                self.conf["tomwer_slices"] = slice

                sub_confs = nabu.interpret_tomwer_configuration(
                    self.conf, scan=self.scan
                )
                self.assertEqual(len(sub_confs), 3 + 1)
        # check from:to:step
        self.conf["tomwer_slices"] = "2:10:2"
        sub_confs = nabu.interpret_tomwer_configuration(self.conf, scan=self.scan)
        self.assertEqual(len(sub_confs), 5 + 1)

    def testPaganinDB(self):
        """Test that tomwer generate one configuration per paganin db"""
        self.conf["phase"] = {}
        # check if single value
        self.conf["phase"]["delta_beta"] = "100"
        sub_confs = nabu.interpret_tomwer_configuration(self.conf, scan=self.scan)
        # not 1 because return each time the 'None' slice which is
        # the reconstruction of the entire volume
        self.assertEqual(len(sub_confs), 1)

        # check single values or list of values
        for pag_db in ("[100, 200, ]", "(225, 230)", "5;6;"):
            with self.subTest(pag_db=pag_db):
                self.conf["phase"]["delta_beta"] = pag_db
                sub_confs = nabu.interpret_tomwer_configuration(
                    self.conf, scan=self.scan
                )
                self.assertEqual(len(sub_confs), 2)
        # check from:to:step
        self.conf["phase"]["delta_beta"] = "200:1000:100"
        sub_confs = nabu.interpret_tomwer_configuration(self.conf, scan=self.scan)
        self.assertEqual(len(sub_confs), 9)


class TestNabuAndAxis(unittest.TestCase):
    """Test that nabu process and axis process are compatible"""

    def setUp(self) -> None:
        self.dataset = "scan_3_"
        self._root_dir = tempfile.mkdtemp()
        data_test_dir = UtilsTest.getEDFDataset(self.dataset)
        scan_folder = os.path.join(self._root_dir, self.dataset)
        shutil.copytree(data_test_dir, scan_folder)
        self.scan = EDFTomoScan(scan_folder)
        [os.remove(cfg_file) for cfg_file in glob(self.scan.path + os.sep + "*.cfg")]

        # initial configuration
        self.conf = get_nabu_default_config(output_dir=self.scan.path)

        self.nabu_process = nabu.NabuSlices(inputs={"data": self.scan})
        self.nabu_process.set_dry_run(True)
        self.nabu_process.set_configuration(self.conf)
        self.scan._axis_params = AxisRP()
        self.nabu_cfg_folder = os.path.join(
            self.scan.path, nabu_settings.NABU_CFG_FILE_FOLDER
        )
        # make sure no .cfg files already exists
        assert len(glob(self.nabu_cfg_folder + os.sep + "*.cfg")) == 0

    def tearDown(self) -> None:
        shutil.rmtree(self._root_dir)

    def test_output_dir(self):
        self.conf["output"]["location"] = "{scan_basename}/output"
        self.nabu_process.set_configuration(self.conf)
        output_dir = self.scan.scan_basename() + os.sep + "output"
        self.assertFalse(os.path.exists(output_dir))
        self.nabu_process.run()
        self.assertTrue(os.path.exists(output_dir))
        nabu_conf_files = glob(
            output_dir + os.sep + nabu_settings.NABU_CFG_FILE_FOLDER + os.sep + "*.cfg"
        )
        self.assertEqual(len(nabu_conf_files), 1)

    def get_saved_rotation_axis(self, nabu_out_file):
        """return the saved value for 'rotation axis' in the nabu
        configuration file"""
        with open(nabu_out_file, "r") as nabu_file:
            lines = nabu_file
            for line in lines:
                if line.startswith("rotation_axis_position"):
                    return line.replace("\n", "").replace(" ", "").split("=")[-1]
            return None

    def testNoCOR(self):
        """test if axis is executed and no COR are set, nabu won't mind"""
        self.scan._axis_params.set_relative_value(None)
        self.nabu_process.run()

        nabu_conf_files = glob(self.nabu_cfg_folder + os.sep + "*.cfg")
        self.assertEqual(len(nabu_conf_files), 1)
        nabu_rotation_axis_value = self.get_saved_rotation_axis(nabu_conf_files[0])
        self.assertEqual(nabu_rotation_axis_value, "")

    def testCOR(self):
        """test that if cor is computed, cor will be registered for nabu"""
        self.scan._axis_params.set_relative_value(2.3)
        self.nabu_process.run()

        nabu_conf_files = glob(self.nabu_cfg_folder + os.sep + "*.cfg")
        self.assertEqual(len(nabu_conf_files), 1)
        nabu_rotation_axis_value = self.get_saved_rotation_axis(nabu_conf_files[0])
        self.assertEqual(
            nabu_rotation_axis_value,
            str(self.scan._axis_params.value_ref_tomwer + self.scan.dim_1 / 2.0),
        )


def get_nabu_default_config(output_dir):
    """:return a default configuration for nabu"""
    return {
        "phase": {
            "padding_type": "edge",
            "method": "",
            "delta_beta": "100.0",
            "margin": 50,
            "unsharp_sigma": 0,
            "unsharp_coeff": 0,
        },
        "preproc": {
            "ccd_filter_threshold": 0.04,
            "take_logarithm": True,
            "log_min_clip": 1e-06,
            "ccd_filter_enabled": 0,
            "log_max_clip": 10.0,
            "flatfield_enabled": 1,
        },
        "dataset": {
            "project_name": "hair_A1_50nm_tomo3_1_",
            "location": "/tmp_14_days/payno/hair_A1_50nm_tomo3_1_",
            "flat_file_prefix": "refHST",
            "projections_subsampling": 1,
            "binning": 1,
            "dark_file_prefix": "dark.edf",
            "binning_z": 1,
        },
        "reconstruction": {
            "positivity_constraint": 1,
            "start_z": 0,
            "end_y": -1,
            "optim_algorithm": "chambolle-pock",
            "enable_halftomo": 0,
            "end_z": -1,
            "iterations": 200,
            "start_x": 0,
            "fbp_filter_type": "ramlak",
            "angle_offset": 0.0,
            "end_x": -1,
            "method": "FBP",
            "angles_file": "",
            "rotation_axis_position": "",
            "start_y": 0,
            "preconditioning_filter": 1,
            "padding_type": "zeros",
            "weight_tv": 0.01,
            "translation_movements_file": "",
            "axis_correction_file": "",
        },
        "output": {"file_format": "hdf5", "location": output_dir},
    }



# coding: utf-8
# /*##########################################################################
#
# Copyright (c) 2017 European Synchrotron Radiation Facility
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# ###########################################################################*/

__authors__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "28/04/2020"


import tempfile
import shutil
import time
import os
from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt
from tomwer.test.utils import skip_gui_test
from tomwer.core.utils.scanutils import MockBlissAcquisition
from orangecontrib.tomwer.widgets.control.DataListenerOW import DataListenerOW
from tomwer.core.process.control.datalistener.rpcserver import BlissAcquisition
from tomwer.core import settings
from glob import glob
import pytest


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestDataListenerSynchronization(TestCaseQt):
    """Insure synchronization is working with the orangecontrib widget"""

    def setUp(self):
        super().setUp()
        self._src_dir = tempfile.mkdtemp()
        self._dst_dir = tempfile.mkdtemp()
        # mock lbsram
        settings.mock_lsbram(True)
        settings._set_lbsram_path(self._src_dir)
        settings._set_dest_path(self._dst_dir)

        scan_path = os.path.join(self._src_dir, "scan0000")
        self.mock_scan = MockBlissAcquisition(
            n_sample=1,
            n_sequence=1,
            n_scan_per_sequence=3,
            n_darks=2,
            n_flats=2,
            output_dir=scan_path,
        )

        self.bliss_acquisition = BlissAcquisition(
            file_path=self.mock_scan.samples[0].sample_file,
            proposal_file=self.mock_scan.proposal_file,
            sample_file=self.mock_scan.samples[0].sample_file,
            entry_name="self.entry = '/1'",
            start_time=time.ctime(),
        )

        self.widget = DataListenerOW()

    def tearDown(self):
        self.mock_scan = None
        self.bliss_acquisition = None
        # self.widget.activate(False)
        self.widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.widget.close()
        self.widget = None
        shutil.rmtree(self._src_dir)
        shutil.rmtree(self._dst_dir)
        super().tearDown()

    def testSynchronization(self):
        """Make sure the datalistener will launch a synchronization on the
        data dir when receive it"""
        self.assertEqual(len(os.listdir(self._dst_dir)), 0)
        self.widget.activate(True)

        self.widget.get_listening_thread()._rpc_sequence_started(
            saving_file=self.bliss_acquisition.sample_file,
            scan_title="toto",
            sequence_scan_number="1",
            proposal_file=self.bliss_acquisition.proposal_file,
            sample_file=self.bliss_acquisition.sample_file,
        )
        time.sleep(1.0)
        self.assertEqual(self.widget._get_n_scan_observe(), 1)
        self.assertEqual(len(os.listdir(self._dst_dir)), 1)
        self.widget.get_listening_thread()._rpc_sequence_ended(
            saving_file=self.bliss_acquisition.sample_file,
            sequence_scan_number="1",
            success=True,
        )
        nx_pattern = os.path.join(self.mock_scan.samples[0].sample_directory, "*.nx")
        self.assertEqual(len(glob(nx_pattern)), 0)
        for i in range(8):
            self.qapp.processEvents()
            time.sleep(1.0)
        # check .nx file has been created
        #
        self.assertEqual(len(glob(nx_pattern)), 1)



# /*##########################################################################
# Copyright (C) 2016-2017 European Synchrotron Radiation Facility
#
# This file is part of the PyMca X-ray Fluorescence Toolkit developed at
# the ESRF by the Software group.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
#############################################################################*/

__author__ = ["H. Payno"]
__license__ = "MIT"
__date__ = "16/08/2021"

import os
from silx.gui import qt
from silx.utils.deprecation import deprecated
from tomwer.core.process.reconstruction.ftseries.ftseries import _Ftseries
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.process.reconstruction.ftseries.params.fastsetupdefineglobals import (
    FastSetupAll,
)
from tomwer.core.process.reconstruction.ftseries import Ftseries
from tomwer.gui.reconstruction.ftserie.h5editor.h5filedialog import H5FileDialog
from tomwer.synctools.ftseries import QReconsParams
from tomwer.synctools.stacks.reconstruction.ftseries import ReconstructionStack
from tomwer.core.scan.scanfactory import ScanFactory
from processview.core.superviseprocess import SuperviseProcess
from tomwer.core.utils import ftseriesutils
import logging
import weakref

_logger = logging.getLogger(__name__)


class FtserieWidget(qt.QWidget, SuperviseProcess):
    _sizeHint = qt.QSize(700, 600)

    sigScanReady = qt.Signal(TomwerScanBase)

    def __init__(self, parent=None, recons_params=None, process_id=None):
        """Widget which read the .hdf5 generated by octave and modify it.
        Then run a subprocess to call octave and run ftseries"""
        qt.QWidget.__init__(self, parent)
        SuperviseProcess.__init__(self, process_id)

        self._explore_for_h5_file = False
        self._dry_run = False
        self._recons_params = recons_params or QReconsParams()
        self._reconstruction_stack = ReconstructionStack(process_id=process_id)
        self._reconstruction_stack.sigReconsFinished.connect(self._notifyScanReady)

        self.setLayout(qt.QVBoxLayout())

        self._recParamSetEditor = None
        self.__buildGUI()
        self._useThread = True

        self._validateButton.hide()
        self._excuteButton = qt.QPushButton("execute", parent=self)
        self._excuteButton.pressed.connect(self._callbackRun)
        self.layout().addWidget(self._excuteButton)

    @property
    def reconstruction_stack(self):
        return self._reconstruction_stack

    @property
    def recons_params(self):
        return self._recons_params

    def setScan(self, scan):
        self._scan = weakref.ref(scan)
        self._recParamSetEditor.setScan(scan)

    @property
    def scan(self):
        if self._scan is None or self._scan() is None:
            return None
        else:
            return self._scan()

    def set_recons_params(self, recons_param):
        _Ftseries.set_recons_params(self, recons_param=recons_param)
        self.getReconsParamSetEditor().setReconsParams(recons_param)

    def _callbackRun(self):
        if isinstance(self._scan, weakref.ref):
            self.pathReceived(self._scan())
        else:
            self.pathReceived(self._scan)

    def isReconsParamsSaved(self):
        """

        :return: True if some previous reconstruction parameters have been
                 saved previously.
        """
        return not (self._reconsParams.params == {})

    def __buildGUI(self):
        """Main function to call the GUI"""
        self.layout().addWidget(self.getReconsParamSetEditor())
        self.layout().addWidget(self.__buildControl())

    def __buildControl(self):
        widget = qt.QWidget(self)
        widget.setLayout(qt.QVBoxLayout())
        widget.layout().setContentsMargins(0, 0, 0, 0)

        widgetLoadSave = qt.QWidget(self)
        widgetLoadSave.setLayout(qt.QHBoxLayout())
        widgetLoadSave.layout().setContentsMargins(0, 0, 0, 0)

        self._loadButton = qt.QPushButton("Load")
        self._loadButton.setToolTip(
            "Will load reconstruction parameters existing in the given file"
        )
        style = qt.QApplication.style()
        self._loadButton.setIcon(style.standardIcon(qt.QStyle.SP_FileIcon))
        self._loadButton.pressed.connect(self.askUserAndLoad)
        widgetLoadSave.layout().addWidget(self._loadButton)
        self._loadButton.setShortcut(qt.QKeySequence.Open)

        spacer = qt.QWidget(self)
        spacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum)
        widgetLoadSave.layout().addWidget(spacer)

        self._saveAs = qt.QPushButton("Save as")
        self._saveAs.setToolTip(
            "Store the current reconstruction parameters into a .h5 file"
        )
        style = qt.QApplication.style()
        self._saveAs.setIcon(style.standardIcon(qt.QStyle.SP_DialogSaveButton))
        self._saveAs.pressed.connect(self._saveAsCallBack)
        widgetLoadSave.layout().addWidget(self._saveAs)
        self._saveAs.setShortcut(qt.QKeySequence.Save)

        widgetValidateCancel = qt.QWidget(self)
        widgetValidateCancel.setLayout(qt.QHBoxLayout())
        widgetValidateCancel.layout().setContentsMargins(0, 0, 0, 0)

        spacer = qt.QWidget(self)
        spacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum)
        widgetValidateCancel.layout().addWidget(spacer)

        self._validateButton = qt.QPushButton("Validate")
        style = qt.QApplication.style()
        self._validateButton.setIcon(style.standardIcon(qt.QStyle.SP_DialogApplyButton))
        self._validateButton.pressed.connect(self._validated)
        widgetValidateCancel.layout().addWidget(self._validateButton)

        self._cancelButton = qt.QPushButton("Cancel")
        self._cancelButton.setIcon(style.standardIcon(qt.QStyle.SP_DialogCancelButton))
        self._cancelButton.pressed.connect(self.__canceled)
        widgetValidateCancel.layout().addWidget(self._cancelButton)

        widget.layout().addWidget(widgetLoadSave)
        widget.layout().addWidget(widgetValidateCancel)
        # the cancel action is undefined so we don't wan't to show this button
        self._cancelButton.setVisible(False)

        return widget

    def getReconsParamSetEditor(self):
        if self._recParamSetEditor is None:
            from tomwer.gui.reconstruction.ftserie.reconsparamseditor import (
                ReconsParamSetEditor,
            )  # avoid cyclic import

            self._recParamSetEditor = ReconsParamSetEditor(
                parent=None, reconsparams=self.recons_params
            )
            self._recParamSetEditor.indexOf(self._recParamSetEditor._dkRefWidget)
            self._reconsParamChanged()
            assert hasattr(self.recons_params, "sigChanged")
            self.recons_params.sigChanged.connect(self._reconsParamChanged)
        return self._recParamSetEditor

    def _reconsParamChanged(self):
        assert self._recParamSetEditor is not None
        self._recParamSetEditor.loadReconsParams(self.recons_params)

    def _updateh5Editor(self, b):
        if self._scan and self._scan.h5File:
            self.reload()

    def askUserH5File(self, save=False):
        """
        Ask user path to a h5File

        :param save: True if the file is used to saved the data. If False then
                     try to open one.
        """
        return H5FileDialog.askForH5File(save=save)

    def _saveAsCallBack(self):
        f = H5FileDialog.askForH5File(save=True)
        if f is not None:
            self.save(h5File=f, displayInfo=True)

    def showValidationButtons(self):
        """Set the buttons for reconstruction parameters visible"""
        self._validateButton.setVisible(True)
        self._cancelButton.setVisible(True)

    def hideValidationButtons(self):
        """Set the buttons for reconstruction parameters not visible"""
        self._validateButton.setVisible(False)
        self._cancelButton.setVisible(False)

    def hideExecuteButton(self):
        self._excuteButton.hide()

    def __canceled(self):
        self.reject()

    def _validated(self):
        self.accept()

    def sizeHint(self):
        """Return a reasonable default size"""
        return self._sizeHint

    def process(self, scan):
        """Call the core function 'run_reconstruction' of the ftseries script
        which will call octave to process reconstruction
        """
        if scan is None:
            return
        self._reconstruction_stack.add(
            scan=scan,
            inputs={
                "pyhst2_params": self.recons_params.to_dict(),
                "explore_for_h5_file": self._explore_for_h5_file,
                "dry_run": self._dry_run,
            },
        )

    def askUserAndLoad(self):
        """Process launch by activing the Load button"""
        f = self.askUserH5File()
        if f is not None:
            self.load_recons_params(f)

    @deprecated(replacement="setScan", since_version="0.8")
    def updatePath(self, path):
        # load if requested for some h5 with reconstruction parameters
        if self._explore_for_h5_file:
            h5file = ftseriesutils.tryToFindH5File(path, "newest")
            "h5 file is the file to load to update the reconstruction parameters"
            if h5file is not None:
                try:
                    self.load_recons_params(h5file)
                except Exception:
                    _logger.warning(
                        "Fail to load reconstruction parameters " "from %s" % h5file
                    )
                else:
                    _logger.info("Reconstruction parameters loaded from " "%s" % h5file)

        if isinstance(path, str):
            self.setScan(ScanFactory.create_scan_object(scan_path=path))
        elif isinstance(path, TomwerScanBase):
            self.setScan(path)
        else:
            raise TypeError(
                f"{path} is expected to be a string or an instance of {TomwerScanBase}"
            )

    def load_recons_params(self, params):
        if isinstance(params, dict):
            self._recons_params.copy(params)
        elif isinstance(params, str):
            if not os.path.isfile(params):
                raise ValueError(f"{params} is not a file")

            fsdg = FastSetupAll()
            fsdg.readAll(params, 3.8)
            self._recons_params.copy(fsdg.structures)
        else:
            raise TypeError(f"{params} is expected to be a file path or a dictionary")

    def save(self, h5File, displayInfo=True):
        Ftseries.save_recons_params(
            recons_params=self.recons_params.to_dict(),
            h5File=h5File,
            displayInfo=displayInfo,
        )

    def setH5Exploration(self, explore_for_h5_file):
        self._explore_for_h5_file = explore_for_h5_file

    def setDryRun(self, dry_run):
        self._dry_run = dry_run

    def _notifyScanReady(self, scan):
        if scan is None:
            return
        if not isinstance(scan, TomwerScanBase):
            raise TypeError(f"{scan} should be None or an instance of {TomwerScanBase}")
        else:
            self.sigScanReady.emit(scan)
